package com.interdataworking.gateway;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.util.*;
import edu.stanford.db.rdf.query.util.Skolem;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;

/**
 * By default, all statements in the body of the rule are discarded
 *
 * RULE 1:
 *
 * (X, property, O_i), ((X, property, O_i), RDFX.order, i)
 * =>
 * (X, property, Y), (Y, type, Seq), Y = Skolem(X, property)
 * (Y, _i, O_i)
 *
 * RULE 2:
 *
 * (O_i, property, X), ((O_i, property, X), RDFX.backwardOrder, i)
 * =>
 * (Y, property, X), (Y, type, Seq), Y = Skolem(X, property)
 * (Y, _i, O_i)
 */

public class OrderReification2RDF extends SimpleModelGateway {

  // we don't want to generate a sequence if only one element is present
  //  HashMap saved = new HashMap(); // Skolem -> original Statement

  public void init(Model srcMD, Model destMD) {
  }

  public Model translate(Model src) throws ModelException {

    Model res = super.translate(src);
    res = new OrderRDF2Reification().translate(res);
    return super.translate(res);
  }

  protected void processStatement(Statement st) throws ModelException {

    applyRule1(st);
    applyRule2(st);
  }
  
  boolean applyRule1(Statement st) throws ModelException {

    if(!RDFX.order.equals(st.predicate()))
      return false;

    Statement os = (Statement)st.subject();

    if(!dest.contains(os)) // sanity check
      return false;

    int i = 1;
    try {
      i = Integer.parseInt(st.object().getLabel());
    } catch (NumberFormatException nfe) {
      return false;
    }

    // remove original statements
    dest.remove(st);
    dest.remove(os);

    // this is resource corresponding to Seq instance
    Resource seq = getUnique(new Skolem(os.subject(), os.predicate()));
    dest.add(stmt(os.subject(), os.predicate(), seq));
    dest.add(stmt(seq, RDF.type, RDF.Seq));
    dest.add(stmt(seq, factory().createOrdinal(i), os.object()));
    
    return true;
  }


  boolean applyRule2(Statement st) throws ModelException {

    if(!RDFX.backwardOrder.equals(st.predicate()))
      return false;

    Statement os = (Statement)st.subject();

    if(!dest.contains(os)) // sanity check
      return false;

    int i = 1;
    try {
      i = Integer.parseInt(st.object().getLabel());
    } catch (NumberFormatException nfe) {
      return false;
    }

    // remove original statements
    dest.remove(st);
    dest.remove(os);

    // this is resource corresponding to Seq instance
    Resource seq = getUnique(new Skolem(os.predicate(), os.object()));
    dest.add(stmt(seq, os.predicate(), os.object()));
    dest.add(stmt(seq, RDF.type, RDF.Seq));
    dest.add(stmt(seq, factory().createOrdinal(i), os.subject()));
    
    return true;
  }

}
