package com.interdataworking.gateway;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.util.*;
import edu.stanford.db.rdf.query.util.Skolem;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

/**
 * By default, all statements in the body of the rule are discarded
 *
 * RULE 1:
 *
 * ((S, ?, ?) | (?, ?, S)) & S = (s,p,o)
 * =>
 * id = Skolem(S)
 * (id, ?, ?)
 * (id, RDF.type, RDF.Statement)
 * (id, RDF.subject, s)
 * (id, RDF.predicate, p)
 * (id, RDF.object, o)
 *
 */

public class Reification2Quad extends SimpleModelGateway {

  public void init(Model srcMD, Model destMD) {
  }

  Resource replace(Statement st) throws ModelException {

    Resource id = getUnique(new Skolem(st));
    dest.add(stmt(id, RDF.type, RDF.Statement));
    dest.add(stmt(id, RDF.subject, st.subject()));
    dest.add(stmt(id, RDF.predicate, st.predicate()));
    dest.add(stmt(id, RDF.object, st.object()));
    doAnotherPass();
    return id;
  }

  protected void processStatement(Statement st) throws ModelException {

    Resource s = null, o = null;
    if(st.subject() instanceof Statement) {
      s = replace((Statement)st.subject());
    }
    if(st.object() instanceof Statement) {
      o = replace((Statement)st.object());
    }
    if(s != null || o != null) {
      //      System.err.println("Removing: " + st);
      dest.remove(st);
      Statement t = stmt(s == null ? st.subject() : s,
			 st.predicate(),
			 o == null ? st.object() : o);
      //      System.err.println("Replacement: " + t);
      dest.add(t);
//       if(dest.contains(t))
// 	System.err.println("---Succeeded");
    }
  }
  
}
