package com.interdataworking.gateway.sqlddl;

import java.io.*;
import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import org.w3c.rdf.syntax.*;
import com.interdataworking.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
//import org.w3c.rdf.vocabulary.dublin_core_19990702.DC;
//import edu.stanford.db.rdf.query.orderings.*;
import edu.stanford.db.rdf.util.PowerModel;

import com.interdataworking.gateway.xml.*;

public class XSD2SQLDDL implements UntypedGateway {

  Obj2SQLDDL os = new Obj2SQLDDL();
  XMLVocabulary XML = new XMLVocabulary();
  XSDVocabulary XSD = new XSDVocabulary();
  DependenciesXSD dx = new DependenciesXSD();

  int fieldCount = 0; // statistics
  int tblIdx = 0;
  Map tables = new HashMap(); // table name (string) -> resource

  Map typeRef = new HashMap(); // field -> String (table name, table resource, or xsd:type)
  Map tablePK = new HashMap(); // resource -> PK field

  Map simpleTypeDef = new HashMap(); // type name (String) -> base (String)

  static Set reservedWords;

  static {
    reservedWords = new HashSet();
    reservedWords.add("create");
    reservedWords.add("table");
    reservedWords.add("references");
    reservedWords.add("unique");
    reservedWords.add("key");
    reservedWords.add("primary");
    reservedWords.add("not");
    reservedWords.add("null");
    reservedWords.add("default");
  }

  public static String replaceReserved(String s) {

    if(s == null)
      return null;

    if(reservedWords.contains(s.toLowerCase()))
      return s + "_";
    else
      return s;
  }


  /*
  void add(Model result, Resource s, Resource p, RDFNode o) throws ModelException {

    result.add(result.getNodeFactory().createStatement(s, p, o));
  }

  void addChild(Resource parent, Resource child, Resource tagName, int pos, Model m) throws ModelException {

    Statement st = m.getNodeFactory().createStatement(parent, XML.child, child);
    m.add(st);
    add(m, child, XML.tagName, tagName);
    add(m, st, RDFX.order, m.getNodeFactory().createLiteral(String.valueOf(pos)));
  }

  void addAttribute(Resource node, Resource attName, String attValue, Model m) throws ModelException {

    add(m, node, attName, m.getNodeFactory().createLiteral(attValue));
  }
  */

  void addMap(Resource src, Resource dest, Collection map) {

    map.add(new MapPair(src, dest, 1.0));
  }

  Resource getTable(PowerModel result, Literal l, String enclName) throws ModelException {

    Resource table = null;
    String tableName = null;

    if(l == null) { // create new name

      tableName = enclName == null ? "TBL_" + tblIdx++ : enclName; // + "_TBL";

      if(tables.containsKey(tableName))
	tableName = tableName + "_" + tblIdx++;

      table = result.unique();
      result.add(table, RDF.type, os.R.TableDef);
      result.add(table, os.R.name, tableName);

      tables.put(tableName, table); // only to keep track of same names

    } else {

      tableName = l.toString();
      table = (Resource)tables.get(tableName);
      if(table == null) {
	table = result.unique();
	result.add(table, RDF.type, os.R.TableDef);
	result.add(table, os.R.name, replaceReserved(tableName));
	tables.put(tableName, table);
      }
    }

    // add primary key
    Resource pk = result.unique();
    result.add(pk, RDF.type, os.R.ColDef);
    //    result.add(pk, os.R.name, "pk_id");
    result.add(pk, os.R.name, tableName.substring(0, 1) + "ID"); //"pk_id");
    result.add(table, os.R.column, pk, 1);

    addSQLType(pk, "int", result);

    Resource pkdef = result.unique();
    result.add(pkdef, RDF.type, os.R.PrimaryKey);
    result.add(pkdef, os.R.keyColumn, pk);
    result.add(table, os.R.key, pkdef);

    tablePK.put(table, pk);
    return table;
  }

  Map sqlTypes = new HashMap(); // string -> resource

  void addSQLType(Resource field, String typeStr, PowerModel result) throws ModelException {

    Resource type = (Resource)sqlTypes.get(typeStr);
    if(type == null) {
      sqlTypes.put(typeStr, type = result.unique());
      result.add(type, RDF.type, os.R.TypeDef);
      result.add(type, os.R.name, replaceReserved(typeStr));
    }      
    result.add(field, os.R.type, type);
  }

  // enclElement: element whose type needs to be completed
  void addChildren(Resource r, PowerModel model, PowerModel result, Collection map, Resource enclTable, Resource enclField, String enclSimpleType) throws ModelException {

    String name = null;

    if(dx.isSchema(model, r)) {

      // new schema encountered; create a dummy table?
    }

    else if(dx.isSimpleType(model, r)) {

      Literal lName = dx.getTypeName(model, r);
      if(lName != null)
	enclSimpleType = lName.toString();
    }

    else if(dx.isComplexType(model, r)) {

      // create a new enclosing table
      String enclName = enclField == null ? null : result.objectS(enclField, os.R.name);

      if(enclTable != null) {
	String enclTableName = result.objectS(enclTable, os.R.name);
	enclName = enclName == null ? enclTableName + "_TBL_" + tblIdx++ : enclTableName + "_" + enclName;
      }

      Resource table = getTable(result, dx.getTypeName(model, r),  enclName);
      addMap(r, table, map);

      if(enclField != null) {
	typeRef.put(enclField, table);
      }

      enclTable = table;
      enclField = null;
    }

    else if(XSD.sequence.equals(dx.getTagName(model, r))) {

      // do nothing, get one level deeper
    }

    else if((enclField != null || enclSimpleType != null) &&
	    XSD.restriction.equals(dx.getTagName(model, r))) {

      // get base
      String elType = RDFUtil.getObjectLiteral(model, r, XSD.base);
      if(elType != null && enclField != null)
	typeRef.put(enclField, elType); // forward reference to type

      if(enclSimpleType != null)
	simpleTypeDef.put(enclSimpleType, elType);
    }

    else if(dx.isElement(model, r)) {

      Literal lName = dx.getElementName(model, r);
      if(lName == null) {
	// System.err.println("Ignoring unnamed element definition");
      } else {

	String elName = replaceReserved(lName.toString());
	
	if(enclTable == null)
	  System.err.println("Ignoring element definition w/o enclosing complex type: " + elName);
	else {

	  //	  System.err.println("Appending field " + elName + " to " + enclTable);
	  
	  // add as field definition to enclosing table
	  fieldCount++;

	  Resource field = result.unique();
	  addMap(r, field, map);
	  
	  result.add(field, RDF.type, os.R.ColDef);
	  result.add(field, os.R.name, elName);
	  
	  String elType = replaceReserved(model.objectS(r, XSD.type));
	  if(elType != null)
	    typeRef.put(field, elType); // forward reference to type
	  // else, will be done using enclField reference in children
	  
	  result.addLast(enclTable, os.R.column, field);
	  enclField = field;
	}
      }
    }

    //    List children = dx.getChildren(model, r);
    List children = model.getOrderedList(r, XML.child);

    for(Iterator it = children.iterator(); it.hasNext();) {

      Resource child = (Resource)it.next();
      addChildren(child, model, result, map, enclTable, enclField, enclSimpleType);
    }
  }


  void addTypeRef(PowerModel model, PowerModel result) throws ModelException {

    for(Iterator it = typeRef.entrySet().iterator(); it.hasNext();) {

      Map.Entry e = (Map.Entry)it.next();
      Resource field = (Resource)e.getKey();
      Object val = e.getValue();

      Resource table = null;

      if(val instanceof String) {
	table = (Resource)tables.get(val); // try to resolve, but may be xsd:int...

      } else if(val instanceof Resource)
	table = (Resource)val;

      //      System.err.println("== e=" + e + ", table=" + table);


      if(table == null) { // this is an elementary type

	String n = (String)val;

	// resolve a simple type reference
	String simple = (String)simpleTypeDef.get(n);
	if(simple != null) {
	  //	  System.err.println("Replacing simple type reference: " + n + " -> " + simple);
	  n = simple;
	}

	if(n.startsWith("xsd:"))
	  n = n.substring(4);

	addSQLType(field, n, result);

      } else { // this is a PK/FK reference

	// make FK an int
	addSQLType(field, "int", result);

	// create a reference from enclField to this table
	Resource refC = result.unique();
	result.add(refC, RDF.type, os.R.ReferentialConstraint);
	result.add(table, os.R.referentialConstraint, refC);
	result.add(refC, os.R.uniqueKeyRole, (Resource)tablePK.get(table));
	result.add(refC, os.R.foreignKeyRole, field);
      }
    }
  }


  /*
	  String e = elementType.toLowerCase();
	  if(e.startsWith("int"))
	    elementType = "xsd:int";
	  else if(e.startsWith("string") || e.startsWith("char") || e.startsWith("varchar"))
	    elementType = "xsd:string";
	  else if(e.startsWith("dec") || e.startsWith("double") || e.startsWith("num"))
	    elementType = "xsd:real";
	  else if(e.startsWith("date"))
	    elementType = "xsd:date";

	// default value
	if(itc.get("D") != null)
	  addAttribute(element, XSD.default_, String.valueOf(itc.get("D")), result);
	//	  w.write(" default=\"" + itc.get("D") + "\"");
	
	// NOT NULL
	if(itc.get("CC") != null)
	  addAttribute(element, XSD.use, "required", result);
  */


  public void convert(PowerModel model, PowerModel result, Collection map) throws ModelException {


    Resource r = dx.getRoot(model);

    addChildren(r, model, result, map, null, null, null);
    addTypeRef(model, result);
    System.err.println("Total: " + tables.size() + " table definitions with " + fieldCount + " attributes");
  }


  public List execute(List input) throws ModelException {

    Model src = (Model)input.get(0);

    Set map = new HashSet();
    PowerModel result = PowerModel.create(src.create());

    convert(PowerModel.create(src), result, map);

    ArrayList l = new ArrayList();
    l.add(result.getPeer());
    l.add(MapPair.asModel(src, map));
    return l;
  }
  

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 2; }

  public int getMaxOutputLen() { return 2; }

}
