package com.interdataworking.gateway.sqlview;

import java.util.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import com.interdataworking.*;
import com.interdataworking.gateway.sqlddl.Obj2SQLDDL;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import edu.stanford.db.rdf.query.orderings.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.util.*;

public class Text2Ref implements UntypedGateway {

  Obj2SQLDDL sqlddl = new Obj2SQLDDL();
  PowerModel result;
  Map replacements = new HashMap();

  public static QueryIterator Q_T = new FindIterator("V", SQLViewVocabulary.tableRef, "T", "ST");

  QueryIterator[] Q_A = {
    new FindIterator("FIELDTERM", SQLViewVocabulary.aliasClause, "A", "ST"),
    new FindIterator("FIELDTERM", SQLViewVocabulary.fieldRef, "F", "ST2"),
    new FindIterator("FIELDTERM", RDF.type, SQLViewVocabulary.FieldTerm)
  };

  static QueryIterator[] Q_FROM = {
    new FindIterator("SELECT", SQLViewVocabulary.fromClause, "FROM"),
    new FindIterator("FROM", RDF.type, SQLViewVocabulary.FromClause),
    new FindIterator("FROM", SQLViewVocabulary.inFromClause, "A"),
    new FindIterator("A", SQLViewVocabulary.aliasName, "N"),
    new FindIterator("A", SQLViewVocabulary.tableRef, "T")
  };

  public static QueryIterator[] Q_VIEW_ = {
    new FindIterator("V", RDF.type, SQLViewVocabulary.CreateView),
    new FindIterator("V", SQLViewVocabulary.cvTable, "CVT"),
    new FindIterator("CVT", SQLViewVocabulary.tableRef, "T"),
    new FindIterator("V", SQLViewVocabulary.as, "S")
  };

  public static QueryIterator Q_VIEW = new OrderIterator(new JoinIterator(Q_VIEW_),
							 new Lexicographic("T"));
  public static QueryIterator[] Q_VIEW_FIELDS_ONLY = {
    new FindIterator("V", SQLViewVocabulary.cvField, "CVF", "ST"),
    new FindIterator("CVF", SQLViewVocabulary.fieldRef, "F", "ST2")
  };

  public static QueryIterator Q_VIEW_FIELDS_ONLY_ORDERED =
    new OrderIterator(new JoinIterator(Q_VIEW_FIELDS_ONLY),
		      new ByOrderProperty("ST"));

  public static QueryIterator[] Q_VIEW_FIELDS = {
    new JoinIterator(Q_VIEW_),
    new JoinIterator(Q_VIEW_FIELDS_ONLY)
  };

  public static Set SELECT_PARENT_ARCS = new HashSet();
  static {
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.selectTerm);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.term);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.fromClause);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.whereClause);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.aliasClause);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.inFromClause);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.headTerm);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.conditionExpr);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.whenTerm);
    SELECT_PARENT_ARCS.add(SQLViewVocabulary.thenTerm);
  }

  public static Resource findEnclosingSelect(Resource r, Model m, boolean checkThis) throws ModelException {

    //    System.err.println("=== find enclosing select of " + r);

    Resource select = r;

    while(true) {

      if(select == null || (checkThis && RDFUtil.isInstanceOf(m, select, SQLViewVocabulary.Select)))
	break;

      checkThis = true;
      r = select;
      select = null;
      
      for(Enumeration en = m.find(null, null, r).elements(); en.hasMoreElements();) {
	
	Statement st = (Statement)en.nextElement();
	if(SELECT_PARENT_ARCS.contains(st.predicate())) {
	  select = st.subject();
	  break;
	}
      }
    }
    return select;
  }

  public static NodeIterator resolveAlias(Resource r, Literal aliasName, Model m) throws ModelException {

    //    System.err.println("=== resolve alias " + r);

    boolean checkThis = true;
    do {
      Resource select = findEnclosingSelect(r, m, checkThis);
      checkThis = false;
      //      System.err.println("=== enclosing select of " + r + " is " + select);
      if(select == null)
	return null;
      NodeIterator it = new NodeResultSet(Q_FROM, m);
      it.set("SELECT", select);
      it.set("N", aliasName);

      if(it.next())
	return it;
      else
	r = select;

    } while(r != null);

    return null;
  }

  public static void bailOut(String msg) throws ModelException {

    throw new ModelException(msg);
  }

  public static void bailOut(String msg, RDFNode n, Model m) throws ModelException {

    String dbg = RDFUtil.getObjectLiteral(m, (Resource)n, SQLViewVocabulary.debugInfo);
    if(dbg == null)
      bailOut(msg);
    else
      bailOut(msg + " " + dbg);
  }

  /*
    Map tables = toMap(Q_T, m, "V");

    for(Iterator it = tables.entrySet().iterator(); it.hasNext();) {

      Map.Entry e = (Map.Entry)it.next();
  */

//   public void add(Resource s, Resource p, RDFNode o) throws ModelException {

//     result.add(result.getNodeFactory().createStatement(s, p, o));
//   }

  void replace(Statement oldSt, Resource s, Resource p, RDFNode o) throws ModelException {

    Statement newSt = result.getNodeFactory().createStatement(s, p, o);
    replacements.put(oldSt, newSt);
    result.remove(oldSt);
    result.add(newSt);
  }

  public Model resolve(Model m_, Model src, Model dest) throws ModelException {

    PowerModel m = PowerModel.create(m_);
    this.result = PowerModel.create(m.duplicate());

    // TABLE REFERENCES

    for(NodeIterator it = new NodeResultSet(Q_T, m); it.next(); ) {

      Resource v = (Resource)it.get("V");
      Literal tableName = (Literal)it.get("T");

      Model target = src;
      String targetStr = "Source";
      if(RDFUtil.isInstanceOf(m, v, SQLViewVocabulary.CVTable)) {
	target = dest;
	targetStr = "Destination";
      }

      Resource table = sqlddl.findTable(target, tableName.getLabel());
      if(table == null)
	bailOut(targetStr + " table " + tableName + " not found", v, m);
      else { // replace
	//	System.err.println("--- replacing " + tableName + " by " + table);
	replace((Statement)it.get("ST"), v, SQLViewVocabulary.tableRef, table);
      }
    }

    // FIELD REFERENCES IN CREATE VIEW HEAD

    for(NodeIterator it = new NodeResultSet(Q_VIEW_FIELDS, m); it.next(); ) {

      Resource v = (Resource)it.get("V");
      Literal tableName = (Literal)it.get("T");
      Literal fieldName = (Literal)it.get("F");

      Resource field = sqlddl.findField(dest, tableName.getLabel(), fieldName.getLabel());

      if(field == null)
	bailOut("Field " + fieldName + " of destination table " + tableName + " not found", v, m);

      //      System.err.println("-- Replacing dest field reference " + fieldName + " by " + field);
      replace((Statement)it.get("ST2"), (Resource)it.get("CVF"), SQLViewVocabulary.fieldRef, field);
    }

    // ALIAS REFERENCES

    for(NodeIterator it = new NodeResultSet(Q_A, m); it.next(); ) {

      Resource fieldTerm = (Resource)it.get("FIELDTERM");
      Literal aliasName = (Literal)it.get("A");

      NodeIterator ia = resolveAlias(fieldTerm, aliasName, m);
      if(ia == null || ia.get("A") == null) {
	bailOut("Alias \"" + aliasName + "\" not defined in the FROM clause", fieldTerm, m);
      } else {
	// replace alias reference
	//	System.err.println("--- Replacing alias reference " + aliasName + " by " + ia.get("A"));

	replace((Statement)it.get("ST"), fieldTerm, SQLViewVocabulary.aliasClause, ia.get("A"));

	// replace field reference
	Literal fieldName = (Literal)it.get("F");
	Resource field = sqlddl.findField(src, ia.get("T").getLabel(), fieldName.getLabel());

	if(field == null)
	  bailOut("Field " + fieldName + " of source table " + ia.get("T") + " not found", fieldTerm, m);

	//	System.err.println("-- Replacing src field reference " + fieldName + " by " + field);
	replace((Statement)it.get("ST2"), fieldTerm, SQLViewVocabulary.fieldRef, field);
      }
    }

    // ESTABLISH REFERENCES FROM CV (HEAD) FIELDS TO CORRESPONDING TERMS IN SELECT CLAUSE

    for(NodeIterator iv = new NodeResultSet(Q_VIEW_, m); iv.next();) {

      Resource view = (Resource)iv.get("V");
      Resource select = (Resource)iv.get("S");

      NodeIterator isrc = new NodeResultSet(Q_VIEW_FIELDS_ONLY_ORDERED, m);
      isrc.set("V", view);
      NodeIterator idest = m.getOrderedNI(select, SQLViewVocabulary.selectTerm);
      
      while(isrc.next() && idest.next()) {
	
	Resource target = (Resource)isrc.get("CVF");
	Resource term = (Resource)idest.get("O");
      
	//	System.err.println("--- head field " + target + " -> term " + term);
	result.add(target, SQLViewVocabulary.cvLink, term);
      }
    }

    RDFUtil.replaceResources(result, replacements);
    return result.getPeer();
  }

  /**
   * Transforms an ordered list of models into another list of models
   */
  public List execute(List input) throws ModelException {

    Model m = (Model)input.get(0);
    Model src = (Model)input.get(1);
    Model dest = (Model)input.get(2);

    Model r = resolve(m, src, dest);

    List result = new ArrayList();
    result.add(r);
    return result;
  }

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 3; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
}
