package com.interdataworking.gateway.xml;

// import java.io.*;
// import org.w3c.rdf.model.*;
// import org.w3c.rdf.syntax.*;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;

import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import java.util.*;
import java.io.*;
import edu.stanford.db.xml.util.GenericSerializer;
import edu.stanford.db.rdf.query.orderings.*;
import com.interdataworking.mm.gui.TLineWriter;
import com.interdataworking.*;

public class RDF2XML implements RDFSerializer, UntypedGateway {

  static final String TAB = "    ";

  XMLVocabulary R = new XMLVocabulary();

  QueryIterator[] Q_ROOT_LOCAL = {
    new FindIterator("P", RDF.type, R.XMLARC),
    new FindIterator("P", "N", "E", "S"),
    new FindIterator("S", RDFX.order, "O"),
    new OptionalIterator(new FindIterator("E", R.cdata, "TXT")),
    new OptionalIterator(new FindIterator("P", R.usesNamespace, "NS"))
  };

  QueryIterator[] Q_CHILDREN_ = {
    new FindIterator("P", "N", "E", "S"),
    new FindIterator("S", RDFX.order, "O"),
    new OptionalIterator(new FindIterator("E", R.cdata, "TXT"))
  };
  QueryIterator Q_CHILDREN_LOCAL = new OrderIterator(new JoinIterator(Q_CHILDREN_),
						     new ByOrderProperty("S", RDFX.order));
  
  QueryIterator[] Q_ATTR_LOCAL_ = {
    new FindIterator("P", "ATT", "VAL", "S"),
    new NegatedIterator(new FindIterator("S", RDFX.order, "O"))
  };

  QueryIterator Q_ATTR_LOCAL = new OrderIterator(new JoinIterator(Q_ATTR_LOCAL_),
  						 new Lexicographic("ATT"));

  public RDF2XML() {
    Q_ROOT = Q_ROOT_LOCAL;
    Q_CHILDREN = Q_CHILDREN_LOCAL;
    Q_ATTR = Q_ATTR_LOCAL;
  }

  protected QueryIterator[] Q_ROOT;
  protected QueryIterator Q_CHILDREN, Q_ATTR;

  protected TLineWriter w;
  protected int depth;
  protected Map namespaces;
  protected Model model;

  public void serialize(Model model, Writer out) throws SerializationException, IOException, ModelException {

    //    System.err.println("Num: " + model.find(null, RDF.type, R.TableDef).size());

    this.w = new TLineWriter(out);
    this.depth = 0;
    this.model = model;

    w.write("<?xml version=\"1.0\"?>\n");

    Resource node = null, name = null;
    Literal txtl = null;
    namespaces = new HashMap();

    int nsId = 0;
    NodeIterator it = new NodeResultSet(Q_ROOT, model);

    while(it.next()) {
      if(node == null) {
	node = (Resource)it.get("E");
	name = (Resource)it.get("N");
	txtl = (Literal)it.get("TXT");
      }
      Resource nsr = (Resource)it.get("NS");
      if(nsr != null) {
	String ns = nsr.getLabel();
	//	System.err.println("NAMESPACE: " + ns);
	if(!namespaces.containsKey(ns)) {
	  String shortcut = nsId <= (int)'z' - (int)'a' ? String.valueOf((char)((int)'a' + nsId++)) : "n" + (nsId++ - (int)'z' + (int)'a');
	  namespaces.put(ns, shortcut);
	}
      }
    }

    if(node != null)
      writeElement(node, name, txtl, namespaces.entrySet().iterator());
  }

  void writeElement(Resource node, Resource name, Literal txtl, Iterator nsIt) throws ModelException, IOException {

    String txt = null;
    if(txtl != null)
      txt = txtl.getLabel();

    indent();
    w.write("<");
    writeName(node, name);

    if(nsIt != null) {

      // comes after one next
      while(nsIt.hasNext()) {

	Map.Entry e = (Map.Entry)nsIt.next();
	char quote = GenericSerializer.abbrevQuote( (String)e.getKey() );

	w.write(" xmlns:" + e.getValue() + "=" + quote + e.getKey() + quote);
      }
    }

    // ATTRIBUTES
    NodeIterator attIt = new NodeResultSet(Q_ATTR, model);
    attIt.set("P", node);

    while(attIt.next()) {

      Object attObj = attIt.get("ATT");
      if(RDF.type.equals(attObj) ||
	 R.tagName.equals(attObj) ||
	 R.child.equals(attObj) ||
	 R.cdata.equals(attObj) ||
	 R.usesNamespace.equals(attObj))
	continue;

      //      System.err.println("ATTRIBUTE: " + attIt.get("ATT") + "=" + attIt.get("VAL"));

      Resource attNameR = (Resource)attIt.get("ATT");
      Literal attValR = (Literal)attIt.get("VAL");
      char quote = GenericSerializer.abbrevQuote( attValR.getLabel() );

      w.write(" ");
      writeName(attNameR);
      w.write("=" + quote + attValR.getLabel() + quote);
    }


    if(txt != null) {// text node
      boolean isCDATA = GenericSerializer.isCDATA(txt);
      if(isCDATA) {
	w.write(" xml:space=\"preserve\">");
	GenericSerializer.writeCDATA(txt, w);
      } else {
	w.write(">");
	w.write(txt);
      }
    }
    NodeIterator chIt = new NodeResultSet(Q_CHILDREN, model);
    chIt.set("P", node);

    boolean hasChildren = chIt.next();
    if(!hasChildren && txt == null) {
      w.write("/>");
    } else {

      if(txt == null)
	w.write(">");

      // process children
      if(hasChildren) {

	depth++;

	do {
	  
	  //	w.write("CHILD: " + chIt.get("E") + " ");
	  writeElement((Resource)chIt.get("E"), (Resource)chIt.get("N"), (Literal)chIt.get("TXT"), null);
	  
	} while(chIt.next());
	
	depth--;

	indent();
      }

      w.write("</");
      writeName(name);
      w.write(">");
    }

    w.flush();
  }

  void writeName(Resource r) throws ModelException, IOException {

    writeName(null, r);
  }

  void writeName(Resource node, Resource r) throws ModelException, IOException {

    String abb = (String)namespaces.get(r.getNamespace());
    String name = abb != null ? abb + ":" + r.getLocalName() : r.getLocalName();
    if(node != null)
      w.write(name, node);
    else
      w.write(name);
  }

  void indent() throws IOException {

    w.write('\n');
    for(int i=0; i < depth; i++)
      w.write(TAB);
  }

  void writeNode(Resource node, Resource name, Literal cdata) throws IOException {

    if(node == null || name == null)
      return;

    indent();
  }

  /**
   * Transforms an ordered list of models into another list of models
   */
  public List execute(List input) throws ModelException {

    Model m = (Model)input.get(0);
    StringWriter w = new StringWriter();
    try {
      serialize(m, w);
    } catch (Exception any) {}

    List result = new ArrayList();
    result.add(w.toString());
    result.add(this.w.getLines());

    //    System.err.println("Returning string:\n " + w + ", lines:\n" + lines);

    return result;
  }

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 2; }

  public int getMaxOutputLen() { return 2; }

}
