package com.interdataworking.gateway.xml;

import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import org.w3c.rdf.util.*;
import edu.stanford.db.rdf.syntax.generic.*;
import edu.stanford.db.xml.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;

import org.xml.sax.*;
import org.xml.sax.helpers.*;

import java.util.*;
import java.io.*;
import java.net.*;

public class XML2RDF2 extends GenericXML2RDF {

  XMLVocabulary R;

  class XMLElement extends Element {

    Resource nodeID;
    int childNum;

    public XMLElement() {
      super();
    }
  }

  XMLElement root = new XMLElement();

  public XML2RDF2() {
    super();
  }

  protected Element createElement() {
    return new XMLElement();
  }

  // implements ContentHandler
  public void startDocument() throws SAXException {

    super.startDocument();
    try {
      R = new XMLVocabulary(); // nodeFactory);
      //      root.nodeID = nodeFactory.createUniqueResource();
      //      current = root;
    } catch (Exception any) {
      throw new SAXException(any);
    }
  }

  public void endDocument() throws SAXException {

    try {
      createStatement(root.nodeID, RDF.type, R.XMLDOM);
      Enumeration en = getNamespaces();
      while(en.hasMoreElements()) {
	createStatement(root.nodeID, R.usesNamespace, nodeFactory.createResource((String)en.nextElement()));
      }
    } catch (Exception any) {
      //      any.printStackTrace(System.err);
      throw new SAXException(any);
    }
    super.endDocument();
  }

  public void startElement(String uri, String local, String qname, Attributes attr) throws SAXException {

    _startElement(uri, local, attr);
    // System.out.println("Started: " + current.getName());

    try {
      XMLElement xe = (XMLElement)current;
      xe.nodeID = nodeFactory.createUniqueResource();
      xe.nodeID = nodeFactory.createResource(xe.nodeID.getNamespace(), xe.nodeID.getLocalName() + "-" + xe.getName().getLocalName());
      //      System.err.println("CREATED: " + xe.nodeID);
      XMLElement parent = (XMLElement)current.getParent();

      Resource elName = nodeFactory.createResource(xe.getName().getNamespace(), xe.getName().getLocalName());
      createStatement(xe.nodeID, R.tagName, elName);

      if(parent != null) {
	parent.childNum++;
	Statement st = createStatement(parent.nodeID,
				       R.child,
				       xe.nodeID);
	createStatement(st, RDFX.order, nodeFactory.createLiteral(parent.childNum));
      } else
	root = xe;
      
      for(Iterator it = xe.getAttributes().entrySet().iterator(); it.hasNext();) {

	Map.Entry e = (Map.Entry)it.next();
	QName name = (QName)e.getKey();
	String value = (String)e.getValue();

//       for(Enumeration en = xe.getAttributes(); en.hasMoreElements();) {
	
//       	QName name = (QName)en.nextElement();
// 	String value = current.getAttribute(name);
	
	createStatement(xe.nodeID,
			nodeFactory.createResource(name.getNamespace(), name.getLocalName()),
			nodeFactory.createLiteral(value));
      }
    } catch (Exception any) {
      throw new SAXException(any);
    }
  }

  public void endElement(String uri, String local, String qname)  throws SAXException {

    try {

      XMLElement xe = (XMLElement)current;

      if((xe.getType() & xe.CDATA) != 0)
	createStatement(xe.nodeID,
			R.cdata,
			nodeFactory.createLiteral(xe.getValue()));

      //      System.out.println("Ended: " + current.getName());

    } catch (Exception any) {
      throw new SAXException(any);
    }

    _endElement(uri, local);
  }

  public static void main(String[] args) throws Exception {

    _main(args[0], new XML2RDF2());
  }
}
