package com.interdataworking.mm;

import org.w3c.rdf.model.*;
import java.util.*;
import com.interdataworking.mm.alg.MapPair;
import com.interdataworking.*;

public abstract class AbstractMapFilter implements MapFilter, UntypedGateway {

  protected Model left, right;

  protected abstract NodeFilter getLeftNodeFilter(Model left) throws ModelException;
  protected abstract NodeFilter getRightNodeFilter(Model right) throws ModelException;

  // additional filtering chance
  protected boolean filterOutPair(RDFNode n1, RDFNode n2) throws ModelException {

    return false;
  }

  public void filterMap(Collection map, Model left, Model right) throws ModelException {

    this.left = left;
    this.right = right;

    NodeFilter lf = getLeftNodeFilter(left);
    NodeFilter rf = getRightNodeFilter(right);

    Iterator it = map.iterator();
    while(it.hasNext()) {

      MapPair p = (MapPair)it.next();
      if(lf.filterOutNode(p.getLeftNode()) || rf.filterOutNode(p.getRightNode()) || filterOutPair(p.getLeftNode(), p.getRightNode())) {
	//	System.err.println("REMOVING: " + p);
	it.remove();
      }
    }
  }

  public List execute(List input) throws ModelException {

    Model map = (Model)input.get(0);
    Model left = (Model)input.get(1);
    Model right = (Model)input.get(2);

    List pairs = MapPair.toMapPairs(map);
    filterMap(pairs, left, right);

    //    System.err.println("Remains after filtering: " + pairs.size());

    ArrayList l = new ArrayList();
    Model dest = MapPair.asModel(left.create(), pairs.toArray());
    l.add(dest);
    return l;
  }
  

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 3; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
}
