package com.interdataworking.mm.alg;

import com.interdataworking.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;

//import edu.stanford.db.rdf.util.Aggregator;

/**
 * Given a set of forward and backward arcs, generates a map of
 * <obj, dependent obj>
 **/

public class GenericDependencies implements UntypedGateway {

  public Set getAggProp() {

    Set aggProp = new HashSet();
    aggProp.add(RDF.type);
    return aggProp;
  }

  public Set getInverseAggProp() {

    return new HashSet();
  }

  /**
   * selects all information that is minimal and absolutely necessary
   * for the referenced nodes to make sense
   */
  public List getDependencies(Model model) throws ModelException {

    List result = new ArrayList();

    Set fw = getAggProp();
    Set bw = getInverseAggProp();

    for(Enumeration en = model.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();

      if(fw.contains(st.predicate())) {

	MapPair p = new MapPair(st.subject(), st.object());
	result.add(p);
      }

      if(bw.contains(st.predicate())) {

	MapPair p = new MapPair(st.object(), st.subject());
	result.add(p);
      }
    }

    return result;
  }

  public List execute(List input) throws ModelException {

    Model src = (Model)input.get(0);

    Model result = MapPair.asModel(src, getDependencies(src));

    ArrayList l = new ArrayList();
    l.add(result);
    return l;
  }
  

  public int getMinInputLen() { return 1; }

  public int getMaxInputLen() { return 1; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
  
}
