// TODO:
// added elements / cardinality: ask user? cardinality constraints may not be enough

package com.interdataworking.mm.alg;

import com.interdataworking.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import java.io.*;
import edu.stanford.db.rdf.util.PowerModel;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import com.interdataworking.mm.interpreter.*;

import edu.stanford.db.rdf.query.*;

/*
  NOTICE: need to know cardinality of properties.
  E.g. merged column, says again of which type? (what is was different, should add, ignore, or replace?)
  In case of type, should ignore.
  When column is added to table, table->new column should be added (incrementally)
*/

  // may cut only at composite and aggregate props!
public class Merge implements UntypedGateway {

  //    static final boolean IGNORE_LITERALS = true;

  boolean DEBUG = false;
  boolean LITERALS_AS_ATTRIBUTES = true; // treat literals as inherent, inseparatable attributes of resources

  PowerModel result;
  NodeFactory f;
  //  PowerModel src, dest;

  Set uniqueSubject = new HashSet();
  Set uniqueObject = new HashSet();
  Set dontCut = new HashSet(); // properties: either both s,o mapped, or none!

  //  boolean appendSrcFirst = true;

  public static final int NONE = -1; // cardinality ignored in constraint

  //  Literal NO_ORDER; // constant used for trashing order

  public Merge() throws ModelException {

    /*
    edu.stanford.db.rdf.uml.vocabulary.uml_behavior_20000507.CommonBehavior R1 = new edu.stanford.db.rdf.uml.vocabulary.uml_behavior_20000507.CommonBehavior();
    edu.stanford.db.rdf.uml.vocabulary.uml_state_20000426.FSM R2 = new edu.stanford.db.rdf.uml.vocabulary.uml_state_20000426.FSM();

    addConstraint(new CardinalityConstraint(R2.StateMachine_top, 1, 1));
    addConstraint(new CardinalityConstraint(R2.Transition_source, NONE, 1));
    addConstraint(new CardinalityConstraint(R2.Transition_target, NONE, 1));
    addConstraint(new CardinalityConstraint(R2.Transition_effect, NONE, 1));
    addConstraint(new CardinalityConstraint(R2.Transition_trigger, NONE, 1));
    */
  }

  int computeMaxIdx(PowerModel dest) throws ModelException {

    // determine the smallest order index in dest
    int maxIdx = 0;

    for(Iterator it = dest.iteratorO(null, RDFX.order); it.hasNext();) {
      
      int idx = -1;
      try {
	idx = Integer.parseInt(((Literal)it.next()).getLabel());
	maxIdx = Math.max(idx + 1, maxIdx);
      } catch (NumberFormatException exc) {}
    }
    return maxIdx;
  }

  // make sure that statements with explicitly mapped subjects/objects go first
  /*
  List prioritySorting(Collection c, Set targetNodes) throws ModelException {

    List[] pl = new List[7];
    for(int i=0; i < pl.length; i++)
      pl[i] = new ArrayList();

    for(Iterator it = c.iterator(); it.hasNext();) {

      Statement st = (Statement)it.next();

      int priority = 0; // [0..6]

      // make sure reified stuff goes last
      if(!(st.subject() instanceof Statement))
	priority += 2;
      if(!(st.object() instanceof Statement))
	priority += 1;

      if(targetNodes.contains(st.subject()))
	priority += 2;
      if(targetNodes.contains(st.object()))
	priority += 1;

      pl[priority].add(st);
    }

    List result = new ArrayList();
    for(int i = pl.length; --i >=0;) // starting with top priority
      result.addAll(pl[i]);

    return result;
  }
  */

  // -: this node is to replace, +: this node is replacement target,
  // o: none of the above, *: both source and target

  // negative priorities (e.g. -1) mean that arcs won't get appended at all
  // boolean value false means that if object is literal, the arc will be trashed

  // priority: {o:9, +:0, *:?, -:3}
  // ISSUE: * should not be allowed/possible at all?

  static Replacement[] replPriority = {
     new Replacement("o o",    9,  true),  //  0=0000
     new Replacement("o +",    8,  true),  //  1=0001
     new Replacement("o -",    7,  true),  //  2=0010
     new Replacement("o *",   -1,  true),  //  3=0011
     new Replacement("+ o",    6,  true),  //  4=0100
     new Replacement("+ +",    4,  true),  //  5=0101
     new Replacement("+ -",    3,  true),  //  6=0110
     new Replacement("+ *",   -1,  true),  //  7=0111
     new Replacement("- o",    5,  false), //  8=1000
     new Replacement("- +",    2,  false), //  9=1001
     new Replacement("- -",    1,  false), // 10=1010
     new Replacement("- *",   -1,  false), // 11=1011
     new Replacement("* o",   -1,  true),  // 12=1100
     new Replacement("* +",   -1,  true),  // 13=1101
     new Replacement("* -",   -1,  true),  // 14=1110
     new Replacement("* *",   -1,  true),  // 15=1111
  };
  static Replacement REIFIED = new Replacement("reified", 0, true);

  /*
  static Replacement[] replPriority = {
     new Replacement("o o",  3+3,  true),  //  0=0000
     new Replacement("o +",  3+2,  true),  //  1=0001
     new Replacement("o -",  3+0,  true),  //  2=0010
     new Replacement("o *",  3+1,  true),  //  3=0011
     new Replacement("+ o",  2+3,  true),  //  4=0100
     new Replacement("+ +",  2+2,  true),  //  5=0101
     new Replacement("+ -",  2+0,  true),  //  6=0110
     new Replacement("+ *",  2+1,  true),  //  7=0111
     new Replacement("- o",  0+3,  false), //  8=1000
     new Replacement("- +",   -1,  false), //  9=1001
     new Replacement("- -",   -1,  false), // 10=1010
     new Replacement("- *",   -1,  false), // 11=1011
     new Replacement("* o",  1+3,  true),  // 12=1100
     new Replacement("* +",  1+2,  true),  // 13=1101
     new Replacement("* -",  1+0,  true),  // 14=1110
     new Replacement("* *",  1+1,  true),  // 15=1111
  };
  static Replacement REIFIED = new Replacement("reified", 0, true);
  */

  static class Replacement {

    String label;
    int priority;
    boolean keepLiterals;

    public Replacement(String label, int priority, boolean keepLiterals) {

      this.label = label;
      this.priority = priority;
      this.keepLiterals = keepLiterals;
    }

    public boolean startMinus() {

      return label.startsWith("-");
    }

    public int getPriority() {

      return priority;
    }

    public boolean keepLiterals() {

      return keepLiterals;
    }

    public String toString() {

      return "[" + label + ": " + (priority >= 0 ? String.valueOf(priority) : "trash") + "]";
    }
  }

  static Replacement getReplacementType(Statement st, Map l2r, Set targetNodes) throws ModelException {

    if(st.subject() instanceof Statement || st.object() instanceof Statement)
      return REIFIED;

    int type = 0;

    if(l2r.containsKey(st.subject()))
      type |= 2;
    if(targetNodes.contains(st.subject()))
      type |= 1;

    type <<= 2; // two positions to the left

    if(l2r.containsKey(st.object()))
      type |= 2;
    if(targetNodes.contains(st.object()))
      type |= 1;

    return replPriority[type];
  }

  static boolean isReified(Statement st) throws ModelException {

    return (st.subject() instanceof Statement || st.object() instanceof Statement);
  }

  // returns a new priority collection
  Collection[] addUntouched(Model m, Map l2r, Set targetNodes, int maxIdx, Collection[] priorityList) throws ModelException {

    //    System.err.println("=== Translation map: " + l2r);

    // 1: non-reified first

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();

      if(isReified(st))
	continue; // ignore reified here

      List l = new ArrayList();
      RDFUtil.replaceMult(st, f, l2r, l);

      // determine type of replacement
      Replacement type = getReplacementType(st, l2r, targetNodes);

      if(DEBUG) // true: && type != REIFIED)
 	System.err.println("-- replacement " + type + " of " + st + " --> " + l.get(0));

      if(LITERALS_AS_ATTRIBUTES && !type.keepLiterals() && st.object() instanceof Literal) {
	if(DEBUG)
	  System.err.println("== trashing literal: " + st);
	continue; // ignore literals if subject has been replaced
      }

      if(type.getPriority() >= 0) {

	for(int i=0; i < l.size(); i++) {

	  Statement ns = (Statement)l.get(i);
	  priorityList = addToPriorityList(priorityList, ns, type.getPriority());
	}
      }
    }

    // 2: now, add reified

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();

      if(!isReified(st))
	continue; // ignore reified here

      List l = new ArrayList();
      RDFUtil.replaceMult(st, f, l2r, l);

      // special treatment of RDFX.order
      if(RDFX.order.equals(st.predicate())) {

	// determine type of replacement
	Replacement type = getReplacementType((Statement)st.subject(), l2r, targetNodes);

	for(int i=0; i < l.size(); i++) {

	  Statement ns = (Statement)l.get(i);

	  if(type.startMinus()) { // subject has changed, trash order

	    ns = translateOrder(ns, maxIdx);
	    if(DEBUG)
	      System.err.println("" + maxIdx + ": increased order from " + st + " to " + ns);
	  }
	  priorityList = addToPriorityList(priorityList, ns, type.getPriority());
	}

      } else { // generic treatment

	// determine type of replacement
	Replacement type = REIFIED;
	
	if(type.getPriority() >= 0) {
	  
	  for(int i=0; i < l.size(); i++) {
	    
	    Statement ns = (Statement)l.get(i);
	    priorityList = addToPriorityList(priorityList, ns, type.getPriority());
	  }
	}
      }
    }

    return priorityList;

    /*
	    if(RDFX.order.equals(ns.predicate()) &&
	       !((Statement)ns.subject()).subject().equals(((Statement)st.subject()).subject())) {
	      
	      ns = translateOrder(ns, maxIdx);
	      //	    System.err.println("" + maxIdx + ": increased order from " + st + " to " + ns);
	    }
    */
  }

  public Model getMerge(Model srcM, Model destM, List replacements) throws ModelException {

    List models = new ArrayList();
    models.add(srcM);
    models.add(destM);
    return getMerge(models, replacements);
  }


  public Model getMerge(List modelsM, List replacements) throws ModelException {

    // first, remove all self matches
    /*
    for(Iterator it = replacements.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      if(p.getLeft().equals(p.getRight()))
	it.remove();
    }

    replacements = MapPair.replacementMap(replacements);
    */

    if(DEBUG)
      MapPair.printMap(replacements, System.err);

    int maxIdx = 0;

    List models = new ArrayList();
    for(Iterator it = modelsM.iterator(); it.hasNext();) {
      PowerModel m = PowerModel.create((Model)it.next());
      models.add(m);
//       if(DEBUG)
// 	System.err.println("Model " + m);
      maxIdx = Math.max(maxIdx, computeMaxIdx(m));
    }

//     this.src = PowerModel.create(srcM);
//     this.dest = PowerModel.create(destM);

    // prepare return values
    this.f = ((Model)modelsM.get(0)).getNodeFactory();
    this.result = PowerModel.create(((Model)modelsM.get(0)).create());
    //    this.NO_ORDER = f.createLiteral("<NO ORDER>");

//     this.f = src.getNodeFactory();
//     this.result = PowerModel.create(dest.create());

//     int maxIdxSrc = computeMaxIdx(src);
//     int maxIdxDest = computeMaxIdx(dest);

    //    System.err.println("=== max idx in src : " + maxIdxSrc);
    //    System.err.println("=== max idx in dest: " + maxIdxDest);

    // compute a single, transitively closed replacement map

    Map l2r = MapPair.toJavaSetNodeMap(replacements, false);
    Set targetNodes = new HashSet(); // this node has been explicitly marked as target node

    for(Iterator it = l2r.values().iterator(); it.hasNext();) {

      Set s = (Set)it.next();
      targetNodes.addAll(s);
    }

    // now, do the replacements and sort out untouched and replaced statements

    Collection[] priorityList = new Collection[0];
    for(Iterator it = models.iterator(); it.hasNext();)
      priorityList = addUntouched((PowerModel)it.next(), l2r, targetNodes, maxIdx, priorityList);

//     priorityList = addUntouched(src, l2r, targetNodes, maxIdxDest, priorityList);
//     priorityList = addUntouched(dest, l2r, targetNodes, maxIdxSrc, priorityList);

    List byPriorities = getPriorityList(priorityList);
    appendFrom(byPriorities);

    //    resolveConflicts(result);

    return result.getPeer();
  }

  Map predTriggered = new HashMap(); // predicate -> Collection of Constraints


  public void addConstraints(Collection col) {

    if(col != null)
      for(Iterator it = col.iterator(); it.hasNext();) {
	
	addConstraint((Constraint)it.next());
      }
  }

  public void addConstraint(Constraint c) {

    Collection pred = c.getPredicateTriggers();

    if(pred != null) {

      for(Iterator it = pred.iterator(); it.hasNext();) {
    
	Resource p = (Resource)it.next();
	Collection coll = (Collection)predTriggered.get(p);
	if(coll == null)
	  predTriggered.put(p, coll = new ArrayList());
	coll.add(c);
      }
    }
  }

  public static interface Constraint {

    // predicate that triggers validation of this constraint (or null if not tested)
    public Collection getPredicateTriggers();

    // would the model still be ok if the statement st is added?
    public boolean isSatisfied(Statement st, Model result) throws ModelException;
  }

  public static class CardinalityConstraint implements Constraint {

    int subjMaxCard; // -1: ignored
    int objMaxCard; // -1: ignored

    Collection pgroup; // predicate group: list of OR-ed predicates

    public CardinalityConstraint(Resource predicate, int subjMaxCard, int objMaxCard) {

      this.pgroup = new ArrayList();
      this.pgroup.add(predicate);
      this.subjMaxCard = subjMaxCard;
      this.objMaxCard = objMaxCard;
    }

    public CardinalityConstraint(Collection pgroup, int subjMaxCard, int objMaxCard) {

      this.pgroup = pgroup;
      this.subjMaxCard = subjMaxCard;
      this.objMaxCard = objMaxCard;
    }

    public Collection getPredicateTriggers() {

      return pgroup;
    }

    public int count(Resource subj, RDFNode obj, Model result) throws ModelException {

      int count = 0;

      for(Enumeration en = result.find(subj, null, obj).elements(); en.hasMoreElements();) {

	Statement st = (Statement)en.nextElement();
	if(pgroup.contains(st.predicate()))
	  count++;
      }

      return count;
    }

    public boolean isSatisfied(Statement st, Model result) throws ModelException {

      if(subjMaxCard >= 0) {
	// make sure there is none with some subject already
	if(count(null, st.object(), result) >= subjMaxCard) {
	  //	  System.err.println("-- Violates unique subject: " + st);
	  return false; // skip it!
	}
      }
      if(objMaxCard >= 0) {
	// make sure there is none with some object already
	if(count(st.subject(), null, result) >= objMaxCard) {
	  //	  System.err.println("-- Violates unique object: " + st);
	  return false; // skip it!
	}
      }
      return true; // happy
    }

    public String toString() {

      return "CardinalityConstraint[" +
	(subjMaxCard >= 0 ? String.valueOf(subjMaxCard) : "*") + "," +
	(objMaxCard >= 0 ? String.valueOf(objMaxCard) : "*") + "]";
    }
  }

  boolean satisfiesConstraints(Statement st, Model result) throws ModelException {

    Collection constraints = (Collection)predTriggered.get(st.predicate());
    if(constraints != null)
      for(Iterator it = constraints.iterator(); it.hasNext();) {

	Constraint c = (Constraint)it.next();
	if(!c.isSatisfied(st, result)) {
	  System.err.println("-- Violates " + c + ": " + st);
	  return false;
	}
      }
    return true; // everybody is happy
  }

  void appendFrom(Collection c) throws ModelException {

    for(Iterator it = c.iterator(); it.hasNext();) {

      Statement st = (Statement)it.next();

      //      System.err.println("Trying to append " + st);

      if(!result.contains(st)) {

	// check whether we may add the arc

	if(!satisfiesConstraints(st, result))
	  continue; // do not add this statement
	
	if(RDFX.order.equals(st.predicate()) &&
	   result.contains((Statement)st.subject())) { // might have been rejected; in this case, ignore

	  Statement old = result.find1(st.subject(), RDFX.order, null);
	  
	  // FIXME: still allows two (subj, pred) with same order!? Need to test for st.subject().subject() as well

	  if(old != null) {
	    
	    int to = result.asInt(st.object());
	    int no = result.asInt(old.object());

	    if(to >= no) {
	      if(DEBUG)
		System.err.println("-- Ignoring >= order: " + st);
	      continue; // skip it!
	    } else
	      result.remove(old);
	  }
	}
	//	System.err.println("adding " + st);
	result.add(st);
      }
    }
  }

  Statement translateOrder(Statement st, int maxIdx) throws ModelException {

    if(maxIdx > 0 && RDFX.order.equals(st.predicate())) {
      // append maxIdx to order position!
      Statement or = result.statement(st.subject(),
				      RDFX.order,
				      result.literal(Integer.parseInt(st.object().getLabel()) + maxIdx));
      //      System.err.println("---   increased order: " + or);
      return or;

    } else // just add it
      return st;
  }

  public List getPriorityList(Collection[] c) {

    List result = new ArrayList();
    for(int i = c.length; --i >= 0;) {

      if(c[i] != null)
	result.addAll(c[i]);
    }
    return result;
  }

  public static Collection[] addToPriorityList(Collection[] c, Object obj, int priority) {

    if(c.length <= priority) {

      Collection[] n = new Collection[priority + 1];
      System.arraycopy(c, 0, n, 0, c.length);
      List l = new ArrayList();
      n[priority] = l;
      l.add(obj);
      return n;

    } else {
      
      if(c[priority] == null)
	c[priority] = new ArrayList();
      c[priority].add(obj);
      return c;
    }
  }

  // merge c1 into c2
  public List execute(List input) throws ModelException {

    Model left = (Model)input.get(0);
    //    Model right = (Model)input.get(1);

    Collection constraints = null;

    LanguageCallback cb = ((ParameterList)input).getExecutionContext().getLanguageCallback(left);

    if(cb == null) {
      System.err.println("[Merge] No callback found, no constraints!");

    } else {

      constraints = cb.getConstraints();
      if(constraints == null)
	System.err.println("[Merge] Callback does not implement, no constraints!");
    }

    addConstraints(constraints);

    Model mapModel = (Model)input.get(input.size()-1);

    List product = MapPair.toMapPairs(mapModel);

    List models = new ArrayList();
    for(int i = 0; i < input.size()-1; i++)
      models.add(input.get(i));
    Model res = getMerge(models, product);

    ArrayList l = new ArrayList();
    l.add(res);
    return l;
  }
  

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 4; } // src, dest, map; constraints schema

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }


  public static void orderedExample(boolean invert) throws Exception {

    System.err.println("This is a simple example of Merge w/ ordered arcs. See the source code for details.");

    RDFFactory rf = new RDFFactoryImpl();
    NodeFactory nf = rf.getNodeFactory();

    // create graph/model A
    PowerModel A = PowerModel.create(rf.createModel());
    PowerModel B = PowerModel.create(rf.createModel());

    Resource arc = A.resource("arc");
    Resource param1 = arc;
    Resource param2 = arc;

    A.add(A.resource("SELECT1"), arc, A.resource("FCT1"));
    A.add(A.resource("FCT1"), param1, A.resource("fn"), 1);
    A.add(A.resource("FCT1"), param2, A.resource("ln"), 2);

    B.add(B.resource("SELECT2"), B.resource("arc"), B.resource("+"));
    B.add(B.resource("+"), param1, B.resource("first"), 1);
    B.add(B.resource("+"), param2, B.resource("++"), 2);
    B.add(B.resource("++"), param1, B.resource("'_'"), 1);
    B.add(B.resource("++"), param2, B.resource("last"), 2);

    List map = new ArrayList();
    map.add(new MapPair(A.resource("SELECT1"), B.resource("SELECT2"), +1));
    map.add(new MapPair(A.resource("FCT1"), B.resource("+"), +1));
    map.add(new MapPair(A.resource("fn"), B.resource("first"), -1));
    map.add(new MapPair(A.resource("ln"), B.resource("last"), -1));

    map = MapPair.replacementMap(map);
    if(invert)
      map = new Invert().invert(map);

    Merge merge = new Merge();
    Model res = merge.getMerge(A, B, map);

    rf.createSerializer().serialize(res, new OutputStreamWriter(System.out));
  }


  public static void orderedExample2(boolean invert) throws Exception {

    System.err.println("This is a second example of Merge w/ ordered arcs. See the source code for details.");

    RDFFactory rf = new RDFFactoryImpl();
    NodeFactory nf = rf.getNodeFactory();

    // create graph/model A
    PowerModel A = PowerModel.create(rf.createModel());
    PowerModel B = PowerModel.create(rf.createModel());

    Resource arc = A.resource("arc");
    Resource name = A.resource("name");

    A.add(A.resource("A"), arc, A.resource("A1"), 1);
    A.add(A.resource("A1"), name, A.literal("CustomerID"));
    A.add(A.resource("A"), arc, A.resource("A2"), 2);
    A.add(A.resource("A2"), name, A.literal("Name"));

    A.add(A.resource("B"), arc, A.resource("B1"), 1);
    A.add(A.resource("B1"), name, A.literal("CustomerID"));
    A.add(A.resource("B"), arc, A.resource("B2"), 2);
    A.add(A.resource("B2"), name, A.literal("PONum"));

    List map = new ArrayList();
    map.add(new MapPair(A.resource("A"), B.resource("B")));
    map.add(new MapPair(A.resource("B1"), B.resource("A1")));
    map.add(new MapPair(A.resource("A2"), B.resource("B1")));

    if(invert)
      map = new Invert().invert(map);

    Merge merge = new Merge();
    Model res = merge.getMerge(A, B, map);

    rf.createSerializer().serialize(res, new OutputStreamWriter(System.out));
  }


  public static void unorderedExample(boolean invert) throws Exception {

    System.err.println("This is a simple example of Merge w/ unordered arcs. See the source code for details.");

    RDFFactory rf = new RDFFactoryImpl();
    NodeFactory nf = rf.getNodeFactory();

    // Create two tiny sample graphs used in the ICDE'02 paper.

    // create graph/model A
    PowerModel A = PowerModel.create(rf.createModel());
    PowerModel B = PowerModel.create(rf.createModel());

    Resource arc = A.resource("arc");
    Resource param1 = A.resource("param1");
    Resource param2 = A.resource("param2");

    A.add(A.resource("SELECT1"), arc, A.resource("FCT1"));
    A.add(A.resource("FCT1"), param1, A.resource("fn"), 1);
    A.add(A.resource("FCT1"), param2, A.resource("ln"), 2);

    B.add(B.resource("SELECT2"), B.resource("arc"), B.resource("+"));
    B.add(B.resource("+"), param1, B.resource("first"), 1);
    B.add(B.resource("+"), param2, B.resource("++"), 2);
    B.add(B.resource("++"), param1, B.resource("'_'"), 1);
    B.add(B.resource("++"), param2, B.resource("last"), 2);

    List map = new ArrayList();
    map.add(new MapPair(A.resource("SELECT1"), B.resource("SELECT2"), +1));
    map.add(new MapPair(A.resource("FCT1"), B.resource("+"), +1));
    map.add(new MapPair(A.resource("fn"), B.resource("first"), -1));
    map.add(new MapPair(A.resource("ln"), B.resource("last"), -1));

    map = MapPair.replacementMap(map);
    if(invert)
      map = new Invert().invert(map);

    Merge merge = new Merge();
    merge.uniqueObject.add(param1);
    merge.uniqueObject.add(param2);

    //    merge.appendSrcFirst = true;

    Model res = merge.getMerge(A, B, map);

    rf.createSerializer().serialize(res, new OutputStreamWriter(System.out));
  }

  public static void transitiveExample(boolean invert) throws Exception {

    System.err.println("This is a simple example of Merge w/ transitive replacements. See the source code for details.");

    RDFFactory rf = new RDFFactoryImpl();
    NodeFactory nf = rf.getNodeFactory();

    // create graph/model A
    PowerModel A = PowerModel.create(rf.createModel());
    PowerModel B = PowerModel.create(rf.createModel());

    Resource arc = A.resource("arc");

    A.add(A.resource("SELECT1"), arc, A.resource("A"));

    B.add(B.resource("SELECT2"), B.resource("arc"), B.resource("FCT"));
    B.add(B.resource("FCT"), arc, B.resource("B"));

    List map = new ArrayList();
    map.add(new MapPair(A.resource("SELECT1"), B.resource("SELECT2")));
    map.add(new MapPair(A.resource("A"), B.resource("FCT")));
    map.add(new MapPair(A.resource("B"), B.resource("A")));

    if(invert)
      map = new Invert().invert(map);

    Merge merge = new Merge();
    //    merge.uniqueObject.add(param1);
    //    merge.uniqueObject.add(param2);

    //    merge.appendSrcFirst = true;

    Model res = merge.getMerge(A, B, map);

    rf.createSerializer().serialize(res, new OutputStreamWriter(System.out));
  }

  public static void main(String[] args) throws Exception {

    boolean invert = args.length > 0;
    orderedExample(invert);
    // orderedExample2(invert);
    // unorderedExample(invert);
    //    transitiveExample(invert);
    // FIXME: trash loose stuff: delete hard on everything (incl. reified) with missing dependencies
    // FIXME: A <-> f(B): A <- f, A->B ==> f(B) -> B
    // FIXME: 1:n is non-deterministic (depends which one gets replaced first)
    // FIXME: translateOrder: need try to replace order if possible (if not, append last)
  }
}
