package com.interdataworking.mm.alg;

import java.util.*;
import com.interdataworking.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import edu.stanford.db.alg.ngram.*;

public class NGramMatcher implements UntypedGateway {

  public int getMinInputLen() { return 2; }

  public int getMaxInputLen() { return 4; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }

  public List execute(List input) throws ModelException {

    Model m1 = (Model)input.get(0);
    Model m2 = (Model)input.get(1);

    Collection l1 = new ArrayList();
    RDFUtil.collectLiterals(m1, l1);
    Collection l2 = new ArrayList();
    RDFUtil.collectLiterals(m2, l2);

    int minLen = 2;
    int maxLen = 4;

    if(input.size() > 3) {
      minLen = Integer.parseInt((String)input.get(2));
      maxLen = Integer.parseInt((String)input.get(3));
    }

    Object[] result = edu.stanford.db.alg.ngram.NGramMatcher.join(l1, l2, minLen, maxLen);

    // copy into MapPairs
    List mapPairs = new ArrayList();
    
    for(int i = 0; i < result.length; i++) {

      ElementSimilarity s = (ElementSimilarity)result[i];
      mapPairs.add(new MapPair((Literal)s.getElement1(), (Literal)s.getElement2(), s.getSim()));
    }

    // NOW, FIND IDENTICAL RESOURCES IN BOTH MODELS AND ADD THEM WITH WEIGHT 1.0!

    Set r1 = new HashSet();
    RDFUtil.collectResources(m1, r1);
    Set r2 = new HashSet();
    RDFUtil.collectResources(m2, r2);

    for(Iterator it = r1.iterator(); it.hasNext();) {

      Resource r = (Resource)it.next();
      if(r2.contains(r)) {
	// this resources appears in both models
	mapPairs.add(new MapPair(r, r, 1.0));
      }
    }

    // RETURN
    ArrayList l = new ArrayList();
    Model map = MapPair.asModel(m1.create(), mapPairs.toArray());
    l.add(map);
    return l;
  }

}
