package com.interdataworking.mm.alg;

import com.interdataworking.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import edu.stanford.db.rdf.model.i.*;
import java.util.*;

public class Prover {

  static Random rnd;

  public static void init() {

    try {
      Registry.getDefaultRegistry().setUniquePrefix("");
    } catch (ModelException any) {}

    Random tmp = new Random();
    int seed = tmp.nextInt(10000);
    System.err.println("Random seeded with " + seed);
    rnd = new Random(seed);
  }

  //  public static RDFFactory rf = new org.w3c.rdf.util.RDFFactoryImpl();
  public static NodeFactory nf = Registry.getDefaultRegistry(); // rf.getNodeFactory();

  // m, m1_m, m2_m are empty
  public static void merge(Model m, Mapping m1_m, Mapping m2_m, Model m1, Model m2, Mapping m1_m2) throws ModelException {

    Set usedM1 = new HashSet();
    Set usedM2 = new HashSet();

    for(Iterator it = m1_m2.iterator(); it.hasNext();) {

      Resource r = unique();
      MapPair p = (MapPair)it.next();

      m.add(r);
      m1_m.add(new MapPair(p.getLeft(), r));
      m2_m.add(new MapPair(p.getRight(), r));

      usedM1.add(p.getLeft());
      usedM2.add(p.getRight());
    }

    for(Iterator it = m1.iterator(); it.hasNext();) {

      Resource t = (Resource)it.next();
      if(!usedM1.contains(t)) {

	Resource r = unique();
	m1_m.add(new MapPair(t, r));
	m.add(r);
      }
    }  

    for(Iterator it = m2.iterator(); it.hasNext();) {

      Resource t = (Resource)it.next();
      if(!usedM2.contains(t)) {

	Resource r = unique();
	m2_m.add(new MapPair(t, r));
	m.add(r);
      }
    }
  }

  public static Mapping confluence(Mapping map1, Mapping map2) throws ModelException {

//     if(true)
//       return intersection(map1, map2);

    Mapping result = new Mapping();

    Model domM1 = domain(map1);
    Model domM2 = domain(map2);
    Model rangeM1 = range(map1);
    Model rangeM2 = range(map2);

    for(Iterator it = map1.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      if(map2.contains(p) ||
	 (!domM2.contains(p.getLeft()) && !rangeM2.contains(p.getRight())))
	result.add(p);
    }

    for(Iterator it = map2.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      if(!domM1.contains(p.getLeft()) && !rangeM1.contains(p.getRight()))
	result.add(p);
    }

    return result;
  }

  public static Mapping extract(Mapping map) throws ModelException {

    Mapping result = new Mapping();

    Map table = indistinguishable(map);

    for(Iterator it = table.values().iterator(); it.hasNext();) {

      Set s = (Set)it.next();
      Resource left = unique();
      for(Iterator it_s = s.iterator(); it_s.hasNext();) {

	Resource r = (Resource)it_s.next();
	result.add(new MapPair(left, r));
      }
    }    

    return result;
  }

  // the values are all sets of indistinguishable nodes
  public static Map indistinguishable(Mapping map) throws ModelException {

    Map intr = MapPair.toJavaSetNodeMap(map, true); // r -> {l}

    // each set {l} determines a set of r nodes pointing to it
    Map table = new HashMap(); // {l} - {r}
    for(Iterator it = intr.entrySet().iterator(); it.hasNext();) {
      
      Map.Entry e = (Map.Entry)it.next();
      Object set = e.getValue();
      Set s = (Set)table.get(set); // lookup by set
      if(s == null)
	table.put(set, s = new HashSet());
      s.add(e.getKey());
    }

    return table;
  }

  public static Mapping diff(Model m, Mapping map) throws ModelException {

    Mapping result = new Mapping();

    Map table = indistinguishable(map);

    Set max = new HashSet();
    Set usedM = new HashSet();

    // iterate over all sets of indistinguishable nodes
    for(Iterator it = table.values().iterator(); it.hasNext();) {

      Set s = (Set)it.next();

      usedM.addAll(s);

      if(s.size() == 1)
	continue; // no need to be made distinguishable

      while(s.size() > max.size())
	max.add(unique());

      Iterator it_s = s.iterator();
      Iterator it_max = max.iterator();

      while(it_s.hasNext()) {

	Resource rs = (Resource)it_s.next();
	Resource rmax = (Resource)it_max.next();
	result.add(new MapPair(rmax, rs));
      }
    }  

    for(Iterator it = m.iterator(); it.hasNext();) {

      Resource t = (Resource)it.next();
      if(!usedM.contains(t)) {

	Resource r = unique();
	result.add(new MapPair(r, t));
      }
    }

    return result;
  }

  public static Mapping invert(Mapping map) {

    Mapping result = new Mapping();

    for(Iterator it = map.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      result.add(new MapPair(p.getRight(), p.getLeft()));
    }

    return result;
  }

  public static Model domain(Mapping map) {

    Model result = new Model();

    for(Iterator it = map.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      result.add(p.getLeft());
    }

    return result;
  }

  public static Model range(Mapping map) {

    Model result = new Model();

    for(Iterator it = map.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      result.add(p.getRight());
    }

    return result;
  }

  public static Mapping compose(Mapping map1, Mapping map2) {

    Map intr = MapPair.toJavaSetMap(map2, false); // l -> {(l,r)}

    Mapping result = new Mapping();

    for(Iterator it = map1.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      Set s = (Set)intr.get(p.getRight());
      if(s != null)
	for(Iterator it_s = s.iterator(); it_s.hasNext();) {

	  MapPair r = (MapPair)it_s.next();
	  result.add(new MapPair(p.getLeft(), r.getRight()));
	}
    }

    return result;
  }

  public static class Model extends HashSet {
  }

  public static class Mapping extends HashSet {
  }

  public static Resource unique() throws ModelException {

    //    System.err.println("create " + nf);

    return nf.createUniqueResource();
  }

  public static Resource create(String ns, String name) throws ModelException {

    //    System.err.println("create " + nf);

    return nf.createResource(ns, name);
  }

  /*
    (a, A)
    (b, B1)
    (b, B2)
    (c1, C)
    (c2, C)
    (d1, D1)
    (d1, D2)
    (d2, D1)
  */
    
  public static Mapping crossMapping(String ns1, String ns2) throws ModelException {

    Mapping map = new Mapping();

    // 1:1
    map.add(new MapPair(create(ns1, "a"), create(ns2, "A")));
    // 1:2
    Resource r = create(ns1, "b");
    map.add(new MapPair(r, create(ns2, "B1")));
    map.add(new MapPair(r, create(ns2, "B2")));
    // 2:1
    r = create(ns2, "C");
    map.add(new MapPair(create(ns1, "c1"), r));
    map.add(new MapPair(create(ns1, "c2"), r));

    return map;
  }

  public static Mapping fctMapping(String ns1, String ns2) throws ModelException {

    Mapping map = new Mapping();

    // 1:1
    map.add(new MapPair(create(ns1, "a"), create(ns2, "A")));

    // 2:1
    Resource r = create(ns2, "C");
    map.add(new MapPair(create(ns1, "c1"), r));
    map.add(new MapPair(create(ns1, "c2"), r));

    return map;
  }


  public static Mapping generalMapping(String ns1, String ns2) throws ModelException {

    return generalMapping(ns1, ns2, true);
  }

  public static Mapping generalMapping(String ns1, String ns2, boolean full) throws ModelException {

    Mapping map = crossMapping(ns1, ns2);

    if(full) {
      // 2:2
      Resource l = create(ns1, "d1");
      Resource r = create(ns2, "D1");
      map.add(new MapPair(l, r));
      map.add(new MapPair(l, create(ns2, "D2")));
      map.add(new MapPair(create(ns1, "d2"), r));
    }

    return map;
  }

  public static Mapping generalMapping(String ns1, String ns2, boolean full, int mult) throws ModelException {

    Mapping result = new Mapping();

    for(int i = 0; i < mult; i++) {
      Mapping map = generalMapping(ns1 + i, ns2 + i, full);
      result.addAll(map);
    }

    return result;
  }

  public static String str(Mapping map) throws ModelException {

    Object[] a = map.toArray();
    //    MapPair.sort(a);

    StringBuffer b = new StringBuffer();

    Iterator it = Arrays.asList(a).iterator();
    while(it.hasNext()) {

      MapPair p = (MapPair)it.next();

      b.append("(" + p.getLeft() + "," + p.getRight() + ")");
      if(it.hasNext())
	b.append(',');
    }

    return b.toString();
  }

  // replaces randomly the nodes from m1 by nodes from m2, or 1:1 if no replacement

  public static Mapping randomMap(Model m1, Model m2, boolean total) {

    Mapping result = new Mapping();
    Object[] arr = m2.toArray();

    for(Iterator it = m1.iterator(); it.hasNext();) {

      Resource r = (Resource)it.next();
      Resource repl = r;
      
      if(total || rnd.nextInt(100) <= 50)
	repl = (Resource)arr[rnd.nextInt(arr.length)];
      
      result.add(new MapPair(r, repl));
    }
    return result;
  }

  public static boolean isOneToOne(Mapping map) throws ModelException {

    return
      compose(invert(map), map).equals(id(range(map))) &&
      compose(map, invert(map)).equals(id(domain(map)));
  }

  public static boolean isBijection(Model s, Mapping map, Model d) throws ModelException {

    return
      compose(invert(map), map).equals(id(d)) &&
      compose(map, invert(map)).equals(id(s));
  }

  public static Mapping id(Model m) throws ModelException {

    Mapping result = new Mapping();

    for(Iterator it = m.iterator(); it.hasNext();) {

      Resource r = (Resource)it.next();
      result.add(new MapPair(r, r));
    }

    return result;
  }

  public static Mapping product(Model m1, Model m2) throws ModelException {

    Mapping result = new Mapping();

    for(Iterator it1 = m1.iterator(); it1.hasNext();) {

      Resource l = (Resource)it1.next();

      for(Iterator it2 = m2.iterator(); it2.hasNext();) {

	Resource r = (Resource)it2.next();
	result.add(new MapPair(l, r));
      }
    }

    return result;
  }

  public static Mapping restrictDomain(Mapping s1_d1, Model s1) throws ModelException {

    Mapping result = new Mapping();
    for(Iterator it = s1_d1.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      if(s1.contains(p.getLeft()))
	result.add(p);
    }
    return result;
  }

  public static Mapping restrictRange(Mapping s1_d1, Model s1) throws ModelException {

    Mapping result = new Mapping();
    for(Iterator it = s1_d1.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      if(s1.contains(p.getRight()))
	result.add(p);
    }
    return result;
  }

  // map2 is adjusted and returned
  public static Mapping generateJoined(Mapping map1, Mapping map2, double prob) throws ModelException {

    Model s1 = range(map1);
    Model s2 = domain(map2);

    int links = (int)(s1.size() * s2.size() * prob);

    Object[] arr1 = s1.toArray();
    Object[] arr2 = s2.toArray();

    Map table = new HashMap();
    
    for(int i=0; i<links; i++) {

      Resource r1 = (Resource)arr1[rnd.nextInt(arr1.length)];
      Resource r2 = (Resource)arr2[rnd.nextInt(arr2.length)];
      table.put(r2, r1);
    }

    Mapping m2_m3 = new Mapping();
    
    for(Iterator it = map2.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();
      Resource repl = (Resource)table.get(p.getLeft());
      if(repl != null)
	m2_m3.add(new MapPair(repl, p.getRight()));
      else
	m2_m3.add(p);
    }

    return m2_m3;
  }

  public static void mergeShortcut(double prob) throws Exception {

    // parameters

    Mapping m1_m2 = generalMapping("m1:", "m2:");
    Mapping m2_m3 = generalMapping("M2:", "m3:");
    m2_m3 = generateJoined(m1_m2, m2_m3, prob);

    Model m1 = domain(m1_m2);
    m1.add(create("m1:", "0"));

    Model m2 = range(m1_m2);
    m2.addAll(domain(m2_m3));
    m2.add(create("m2:", "0"));

    Model m3 = range(m2_m3);
    m3.add(create("m3:", "0"));

    // script

    Model m12 = new Model();
    Mapping m1_m12 = new Mapping();
    Mapping m2_m12 = new Mapping();
    merge(m12, m1_m12, m2_m12, m1, m2, m1_m2);

    Model m23 = new Model();
    Mapping m2_m23 = new Mapping();
    Mapping m3_m23 = new Mapping();
    merge(m23, m2_m23, m3_m23, m2, m3, m2_m3);

    Mapping m12_m3 = compose(invert(m2_m12), m2_m3);

    Model ma = new Model();
    Mapping m12_ma = new Mapping();
    Mapping m3_ma = new Mapping();
    merge(ma, m12_ma, m3_ma, m12, m3, m12_m3);

    Mapping m12_m23 = compose(invert(m2_m12), m2_m23);
    
    Model mb = new Model();
    Mapping m12_mb = new Mapping();
    Mapping m23_mb = new Mapping();
    merge(mb, m12_mb, m23_mb, m12, m23, m12_m23);

    Mapping ma_mb = confluence(compose(invert(m12_ma), m12_mb),
			       compose(compose(invert(m3_ma), m3_m23), m23_mb));

    //    System.err.println("ma_mb = " + str(ma_mb));
    boolean bij = isBijection(ma, ma_mb, mb);
    System.err.println("bijection: " + bij);
    if(!bij)
      System.exit(-1);
  }


  public static void mergeAssociative(double prob) throws Exception {

    // parameters

    Mapping m1_m2 = generalMapping("m1:", "m2:");
    Mapping m2_m3 = generalMapping("M2:", "m3:");
    m2_m3 = generateJoined(m1_m2, m2_m3, prob);

    Model m1 = domain(m1_m2);
    m1.add(create("m1:", "0"));

    Model m2 = range(m1_m2);
    m2.addAll(domain(m2_m3));
    m2.add(create("m2:", "0"));

    Model m3 = range(m2_m3);
    m3.add(create("m3:", "0"));

    // script

    Model m12 = new Model();
    Mapping m1_m12 = new Mapping();
    Mapping m2_m12 = new Mapping();
    merge(m12, m1_m12, m2_m12, m1, m2, m1_m2);

    Model m23 = new Model();
    Mapping m2_m23 = new Mapping();
    Mapping m3_m23 = new Mapping();
    merge(m23, m2_m23, m3_m23, m2, m3, m2_m3);

    Mapping m12_m3 = compose(invert(m2_m12), m2_m3);
    Mapping m23_m1 = compose(invert(m2_m23), invert(m1_m2));

    Model ma = new Model();
    Mapping m12_ma = new Mapping();
    Mapping m3_ma = new Mapping();
    merge(ma, m12_ma, m3_ma, m12, m3, m12_m3);
    
    Model mb = new Model();
    Mapping m23_mb = new Mapping();
    Mapping m1_mb = new Mapping();
    merge(mb, m23_mb, m1_mb, m23, m1, m23_m1);

    Mapping ma_mb = confluence(confluence(compose(compose(invert(m12_ma), invert(m1_m12)), m1_mb),
					  compose(compose(compose(invert(m12_ma), invert(m2_m12)), m2_m23), m23_mb)),
			       compose(compose(invert(m3_ma), m3_m23), m23_mb));

    //    System.err.println("ma_mb = " + str(ma_mb));
    boolean bij = isBijection(ma, ma_mb, mb);
    System.err.println("bijection: " + bij);
    if(!bij)
      System.exit(-1);
  }

  public static Mapping union(Mapping map1, Mapping map2) throws ModelException {

    Mapping result = new Mapping();
    result.addAll(map1);
    result.addAll(map2);
    return result;
  }

  public static Mapping intersection(Mapping map1, Mapping map2) throws ModelException {

    Mapping result = new Mapping();
    for(Iterator it = map1.iterator(); it.hasNext();) {

      Object n = it.next();
      if(map2.contains(n))
	result.add(n);
    }
    return result;
  }

  public static void mergeLine2(Model m1, Model m2, Model m3,
			       Mapping m1_m2, Mapping m2_m3, Mapping m1_m3) throws ModelException {

    Model m12 = new Model();
    Mapping m1_m12 = new Mapping();
    Mapping m2_m12 = new Mapping();
    merge(m12, m1_m12, m2_m12, m1, m2, m1_m2);

    Mapping m12_m3 = compose(invert(m2_m12), m2_m3);
    //    Mapping m23_m1 = compose(invert(m2_m23), invert(m1_m2));

    Model ma = new Model();
    Mapping m12_ma = new Mapping();
    Mapping m3_ma = new Mapping();
    merge(ma, m12_ma, m3_ma, m12, m3, m12_m3);

    Mapping m1_ma = compose(m1_m12, m12_ma);

    //    m1_m3 = confluence(m1_m3, compose(m1_m2, m2_m3));
    m1_m3 = compose(m1_m2, m2_m3);

    Mapping ma_self = compose(compose(invert(m1_ma), m1_m3), m3_ma);

    //    ma_self = confluence(ma_self, invert(ma_self));

    Model mA = new Model();
    Mapping ma_mA_1 = new Mapping();
    Mapping ma_mA_2 = new Mapping();
    merge(mA, ma_mA_1, ma_mA_2, ma, ma, ma_self);
    
    //    System.err.println("ma = " + ma + ", size=" + ma.size());
    System.err.println("mA = " + mA + ", size=" + mA.size());
  }


  // not even for total surjective cross-mappings...
  public static void mergeLine(Model m1, Model m2, Model m3,
			       Mapping m1_m2, Mapping m2_m3, Mapping m1_m3) throws ModelException {

    Model m12 = new Model();
    Mapping m1_m12 = new Mapping();
    Mapping m2_m12 = new Mapping();
    merge(m12, m1_m12, m2_m12, m1, m2, m1_m2);

    Mapping m12_m3 = confluence(compose(invert(m2_m12), m2_m3),
				compose(invert(m1_m12), m1_m3));

    Model ma = new Model();
    Mapping m12_ma = new Mapping();
    Mapping m3_ma = new Mapping();
    merge(ma, m12_ma, m3_ma, m12, m3, m12_m3);

    System.err.println("ma = " + ma + ", size=" + ma.size());
  }

  public static void mergeFullyAssociative(double prob) throws ModelException {

    // parameters

//     Mapping m1_m2 = crossMapping("m1a:", "m2a:");
//     Mapping m2_m3 = crossMapping("m2b:", "m3a:");
//     Mapping m1_m3 = crossMapping("m1b:", "m3b:");

    Mapping m1_m2, m2_m3, m1_m3;

    m1_m2 = generalMapping("m1a:", "m2a:", false, 3);
    m2_m3 = generalMapping("m2b:", "m3a:", false, 3);
    m1_m3 = generalMapping("m1b:", "m3b:", false, 3);
    m2_m3 = generateJoined(m1_m2, m2_m3, prob);
    m1_m3 = invert(generateJoined(m2_m3, invert(m1_m3), prob));
    m1_m2 = generateJoined(invert(m1_m3), m1_m2, prob);

    // make all mappings total surjective

    for(int i=0; i < 100; i++) {
      m1_m2 = restrictDomain(m1_m2, domain(m1_m3));
      m1_m3 = restrictDomain(m1_m3, domain(m1_m2));
      m1_m2 = restrictRange(m1_m2, domain(m2_m3));
      m2_m3 = restrictDomain(m2_m3, range(m1_m2));
      m2_m3 = restrictRange(m2_m3, range(m1_m3));
      m1_m3 = restrictRange(m1_m3, range(m2_m3));
    }


    m1_m2 = product(domain(m1_m2), range(m1_m2));
    m2_m3 = product(domain(m2_m3), range(m2_m3));
    m1_m3 = product(domain(m1_m3), range(m1_m3));

    boolean prod = compose(m1_m2, m2_m3).equals(product(domain(m1_m2), range(m2_m3)));

    System.err.println("PRODUCT: " + prod);


    Model m1 = domain(m1_m2);
    m1.addAll(domain(m1_m3));
    //    m1.add(create("m1:", "0"));

    Model m2 = range(m1_m2);
    m2.addAll(domain(m2_m3));
    //    m2.add(create("m2:", "0"));

    Model m3 = range(m2_m3);
    m3.addAll(range(m1_m3));
    //    m3.add(create("m3:", "0"));

    System.err.println("m1_m2 = " + str(m1_m2));
    System.err.println("m2_m3 = " + str(m2_m3));
    System.err.println("m1_m3 = " + str(m1_m3));


    mergeLine(m1, m2, m3, m1_m2, m2_m3, m1_m3);

    mergeLine(m1, m3, m2, m1_m3, invert(m2_m3), m1_m2);

    /*
    // now first join m2 and m3

    Model m23 = new Model();
    Mapping m2_m23 = new Mapping();
    Mapping m3_m23 = new Mapping();
    merge(m23, m2_m23, m3_m23, m2, m3, m2_m3); // confluence(m2_m3, compose(invert(m1_m2), m1_m3)));

    Mapping m23_m1 = compose(invert(m2_m23), invert(m1_m2));

    Model mb = new Model();
    Mapping m23_mb = new Mapping();
    Mapping m1_mb = new Mapping();
    merge(mb, m23_mb, m1_mb, m23, m1, m23_m1);

    Mapping m3_mb = compose(m3_m23, m23_mb);

    Mapping mb_self = compose(compose(invert(m1_mb), m1_m3), m3_mb);

    Model mB = new Model();
    Mapping mb_mB_1 = new Mapping();
    Mapping mb_mB_2 = new Mapping();
    merge(mB, mb_mB_1, mb_mB_2, mb, mb, mb_self);

    //    System.err.println("mb = " + mb + ", size=" + mb.size());
    System.err.println("mB = " + mB + ", size=" + mB.size());
    */
  }

  public static void mergeFullyAssociative2(double prob) throws ModelException {

    // parameters

    Mapping m1_m2 = generalMapping("m1a:", "m2a:");
    Mapping m2_m3 = generalMapping("m2b:", "m3a:");
    Mapping m1_m3 = generalMapping("m1b:", "m3b:");
    m2_m3 = generateJoined(m1_m2, m2_m3, prob);
    m1_m3 = invert(generateJoined(m2_m3, invert(m1_m3), prob));
    m1_m2 = generateJoined(invert(m1_m3), m1_m2, prob);

    //    m1_m3 = compose(m1_m2, m2_m3); // ???

//     m1_m2 = confluence(m1_m2, compose(m1_m3, invert(m2_m3)));
//     m2_m3 = confluence(m2_m3, compose(invert(m1_m2), m1_m3));
//     m1_m3 = confluence(m1_m3, compose(m1_m2, m2_m3));

//     m1_m2 = M1_M2;
//     m2_m3 = M2_M3;
//     m1_m3 = M1_M3;

    Model m1 = domain(m1_m2);
    m1.addAll(domain(m1_m3));
    m1.add(create("m1:", "0"));

    Model m2 = range(m1_m2);
    m2.addAll(domain(m2_m3));
    m2.add(create("m2:", "0"));

    Model m3 = range(m2_m3);
    m3.addAll(range(m1_m3));
    m3.add(create("m3:", "0"));

    System.err.println("m1_m2 = " + str(m1_m2));
    System.err.println("m2_m3 = " + str(m2_m3));
    System.err.println("m1_m3 = " + str(m1_m3));

    // script

    Model m12 = new Model();
    Mapping m1_m12 = new Mapping();
    Mapping m2_m12 = new Mapping();
    merge(m12, m1_m12, m2_m12, m1, m2, m1_m2); // confluence(m1_m2, compose(m1_m3, invert(m2_m3))));

    Model m23 = new Model();
    Mapping m2_m23 = new Mapping();
    Mapping m3_m23 = new Mapping();
    merge(m23, m2_m23, m3_m23, m2, m3, m2_m3); // confluence(m2_m3, compose(invert(m1_m2), m1_m3)));

    Mapping m12_m3 = confluence(compose(invert(m2_m12), m2_m3),
				compose(invert(m1_m12), m1_m3));

    Model ma = new Model();
    Mapping m12_ma = new Mapping();
    Mapping m3_ma = new Mapping();
    merge(ma, m12_ma, m3_ma, m12, m3, m12_m3);

    Mapping m23_m1 = confluence(compose(invert(m2_m23), invert(m1_m2)),
				compose(invert(m3_m23), invert(m1_m3)));
    
    Model mb = new Model();
    Mapping m23_mb = new Mapping();
    Mapping m1_mb = new Mapping();
    merge(mb, m23_mb, m1_mb, m23, m1, m23_m1);

    Mapping ma_mb = confluence(confluence(compose(compose(invert(m12_ma), invert(m1_m12)), m1_mb),
					  compose(compose(invert(m3_ma), m3_m23), m23_mb)),
			       compose(compose(compose(invert(m12_ma), invert(m2_m12)), m2_m23), m23_mb));

    System.err.println("ma_mb = " + str(ma_mb));
    boolean bij = isBijection(ma, ma_mb, mb);
    System.err.println("bijection: " + bij);
    if(!bij) {
      System.err.println("ma = " + ma);
      System.err.println("mb = " + mb);
      System.exit(-1);
    }
  }

  // holds!!!
  public static void doubleExtractConjecture(double prob) throws Exception {

    Mapping q1 = generalMapping("m1:", "s1:", true, 3);
    Mapping q2 = generalMapping("m2:", "s2:", true, 3);
    q2 = generateJoined(invert(q1), q2, prob);

    q1 = restrictDomain(q1, domain(q2)); // same domains
    q2 = restrictDomain(q2, domain(q1));

//     Mapping a_m2 = extract(m1_m2);
//     Mapping b_m2 = extract(m3_m2);

    // both approaches work!!!
    Mapping s1_s2 = compose(invert(q1), q2);
    //  Mapping s1_s2 = product(range(q1), range(q2));

    // empty mapping does not work
    //    Mapping s1_s2 = new Mapping();

    // Mapping a_b = compose(confluence(a_m2, b_m2), invert(confluence(a_m2, b_m2)));
    
    Model s = new Model();
    Mapping s1_s = new Mapping();
    Mapping s2_s = new Mapping();
    merge(s, s1_s, s2_s, range(q1), range(q2), s1_s2);

    Mapping m_s = confluence(compose(q1, s1_s), compose(q2, s2_s));

    Mapping mx_m = extract(invert(m_s));

    Mapping q1_ = compose(compose(invert(mx_m), mx_m), q1);
    Mapping q2_ = compose(compose(invert(mx_m), mx_m), q2);
    boolean eq1 = q1.equals(q1_);
    boolean eq2 = q2.equals(q2_);

    if(!eq1 || !eq2) {
      System.err.println("q1 = " + str(q1));
      System.err.println("q2 = " + str(q2));
      System.err.println("s1_s2 = " + str(s1_s2));
      
      System.err.println("s = " + s);
      System.err.println("s1_s = " + str(s1_s));
      System.err.println("s2_s = " + str(s2_s));
      
      System.err.println("m_s = " + str(m_s));
      System.err.println("mx_m = " + str(mx_m));
      
      System.err.println("q1' = " + str(q1_));
      System.err.println("q1 = " + str(q1));
      
      System.err.println("equal?: " + eq1);
      System.err.println("q2' = " + str(q1_));
      System.err.println("q2 = " + str(q1));
      
      System.err.println("equal?: " + eq2);

      System.exit(1);
    } else
      System.err.println("Both equal");
  }

  public static void doubleExtractMergeTest(double prob) throws Exception {

    Mapping m1_m2 = generalMapping("m1:", "m2a:", true, 1);
    Mapping m2_m3 = generalMapping("m2b:", "m3:", true, 1);
    m2_m3 = generateJoined(m1_m2, m2_m3, prob);
    Mapping m3_m2 = invert(m2_m3);

    m1_m2 = restrictRange(m1_m2, range(m3_m2)); // same ranges!!! works for two views over m2
    m3_m2 = restrictRange(m3_m2, range(m1_m2));

    // counterexample: {x-a, x-b}, {y-b, y-c}

    Mapping a_m2 = extract(m1_m2);
    Mapping b_m2 = extract(m3_m2);

    Mapping a_b = compose(a_m2, invert(b_m2));
    // Mapping a_b = compose(confluence(a_m2, b_m2), invert(confluence(a_m2, b_m2)));
    
    Model P = new Model();
    Mapping a_P = new Mapping();
    Mapping b_P = new Mapping();
    merge(P, a_P, b_P, domain(a_m2), domain(b_m2), a_b);
    Mapping P_m2 = confluence(compose(invert(a_P), a_m2), compose(invert(b_P), b_m2));

    Mapping all_m2 = union(m1_m2, m3_m2); // not confluence!!!
    Mapping Q_m2 = extract(all_m2);

    Model Q = domain(Q_m2);

    Mapping P_Q = compose(P_m2, invert(Q_m2));
    boolean bij = isBijection(P, P_Q, Q);

    System.err.println("bijection: " + bij);

    if(!bij) {
      System.err.println("m1_m2 = " + str(m1_m2));
      System.err.println("m3_m2 = " + str(m3_m2));

      System.err.println("a_b = " + str(a_b));
      System.err.println("P = " + P + ", size=" + P.size());
      System.err.println("P_m2 = " + str(P_m2));

      System.err.println("all_m2 = " + str(all_m2));

      System.err.println("Q = " + Q + ", size=" + Q.size());
      System.err.println("Q_m2 = " + str(Q_m2));

      System.err.println("P_Q = " + str(P_Q));

      System.err.println("one-to-one: " + isOneToOne(P_Q));

      System.exit(1);
    }
  }

  public static void propDelTest() throws Exception {

//     Mapping s1_d1 = generalMapping("s1:", "d1:");
//     Mapping t_s2 = generalMapping("S1:", "s2:");

/*
    Mapping s1_d1 = crossMapping("s1:", "d1:");
    Mapping t_s2 = crossMapping("S1:", "s2:");

    Model dom1 = domain(s1_d1);
    Model dom2 = domain(t_s2);

    Mapping s1_s2 = compose(invert(randomMap(dom2, dom1, true)), t_s2);
*/

    Mapping t_d1 = crossMapping("S1:", "d1:");
    Mapping s1_s2 = crossMapping("s1:", "s2:");

    Model dom1 = domain(t_d1);
    Model dom2 = domain(s1_s2);

    Mapping s1_d1 = compose(invert(randomMap(dom1, dom2, false)), t_d1);

    s1_d1 = restrictDomain(s1_d1, domain(s1_s2));
    s1_s2 = restrictDomain(s1_s2, domain(s1_d1));


    Model s1 = domain(s1_s2);
    s1.addAll(domain(s1_d1));
    s1.add(create("s1:","0"));

    Model d1 = range(s1_d1);
    d1.add(create("d1:","0"));

    System.err.println("s1_d1 = " + str(s1_d1));
    System.err.println("s1_s2 = " + str(s1_s2));

    Mapping s1p_s1 = diff(s1, invert(s1_s2));
    /*
    System.err.println("s1p_s1 = " + str(s1p_s1));
    System.err.println("composed = " + str(compose(compose(invert(s1p_s1), s1p_s1), s1_d1)));
    Mapping cmap = confluence(s1_d1, compose(compose(s1p_s1, invert(s1p_s1)), s1_d1));
    System.err.println("confluence = " + str(cmap));
    //    d1p_d1 = extract(
    */

    Mapping restr = restrictDomain(s1_d1, range(s1p_s1));
    System.err.println("restricted = " + str(restr));

    Mapping d1p_d1_A = extract(restr);
    System.err.println("propagated1 = " + str(d1p_d1_A));

    // alternative
    Mapping s2_d1 = compose(invert(s1_s2), s1_d1);

    Mapping d1d_d1 = diff(d1, s1_d1); // added in d1
    Mapping d1x_d1 = extract(s2_d1); // kept in d1

    Model d1p = new Model();
    Mapping d1x_d1p = new Mapping(), d1d_d1p = new Mapping();
    merge(d1p, d1x_d1p, d1d_d1p, domain(d1x_d1), domain(d1d_d1), compose(d1x_d1, invert(d1d_d1)));

    Mapping d1p_d1_B = confluence(compose(invert(d1d_d1p), d1d_d1),
				  compose(invert(d1x_d1p), d1x_d1));

    System.err.println("propagated2 = " + str(d1p_d1_B));

    compareRange(d1p_d1_A, d1p_d1_B);
  }

  public static void compareRange(Mapping map1, Mapping map2) throws Exception {

    Map ind1 = indistinguishable(map1); 
    Map ind2 = indistinguishable(map2);

    HashSet s1 = new HashSet(); // set of sets
    HashSet s2 = new HashSet();
    s1.addAll(ind1.values());
    s2.addAll(ind2.values());

    s1.removeAll(ind2.values());
    s2.removeAll(ind1.values());

    System.err.println("map1-map2: " + s1);
    System.err.println("map2-map1: " + s2);
  }

  public static void diffTest() throws Exception {

    Mapping mP_m = generalMapping("m':", "m:");
    Model m = range(mP_m);
    m.add(unique());

    System.err.println("m'_m = " + str(mP_m));
    System.err.println("m = " + m);

    Mapping md_m = diff(m, mP_m);

    System.err.println("md_m = " + str(md_m));

    // other way
    Mapping mx_m = extract(mP_m);

    System.err.println("mx_m = " + str(mx_m));

    Model mo = new Model();
    Mapping mx_mo = new Mapping();
    Mapping md_mo = new Mapping();

    Mapping mx_md = compose(mx_m, invert(md_m));
    //    System.err.println("mx_md = " + str(mx_md));

    merge(mo, mx_mo, md_mo, domain(mx_m), domain(md_m), mx_md);

    System.err.println("md_mo = " + str(md_mo));

    System.err.println("m is " + m);
    System.err.println("mo is " + mo);
  }

  public static void main(String[] args) throws Exception {

    Prover.init();

    //    doubleExtractConjecture(0.5);

    //    doubleExtractMergeTest(0.5);
    //    mergeFullyAssociative(0.1);

    for(int i=0; i < 5000; i++) {
      double prob = Prover.rnd.nextDouble();
      doubleExtractConjecture(0.5);
//      doubleExtractMergeTest(prob);
      //       mergeShortcut(prob);
      //      mergeAssociative(prob);
    }


    //    propDelTest();
    //    diffTest();

    /*
    NodeFactory nf = Prover.nf;

    Mapping map = new Mapping();
    map.add(new MapPair(nf.createResource("a"), nf.createResource("1")));
    map.add(new MapPair(nf.createResource("b"), nf.createResource("1")));
    map.add(new MapPair(nf.createResource("c"), nf.createResource("2")));
    map.add(new MapPair(nf.createResource("d"), nf.createResource("3")));
    map.add(new MapPair(nf.createResource("e"), nf.createResource("3")));
    map.add(new MapPair(nf.createResource("f"), nf.createResource("3")));

    //    Mapping result = Prover.extract(map);
    Mapping result = Prover.diff(map);
    System.err.println("Result: " + result);
    System.err.println("Domain: " + domain(result));
    */

    /*
    Mapping map1 = new Mapping();
    map1.add(new MapPair(nf.createResource("1"), nf.createResource("a")));
    map1.add(new MapPair(nf.createResource("1"), nf.createResource("b")));
    map1.add(new MapPair(nf.createResource("2"), nf.createResource("c")));

    Mapping map2 = new Mapping();
    map2.add(new MapPair(nf.createResource("a"), nf.createResource("A1")));
    map2.add(new MapPair(nf.createResource("a"), nf.createResource("A2")));
    map2.add(new MapPair(nf.createResource("c"), nf.createResource("C")));

    Mapping map3 = Prover.compose(map1, map2);
    System.err.println("Result: " + map3);
    */
  }
}
