package com.interdataworking.mm.alg;

import com.interdataworking.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import java.io.*;
import edu.stanford.db.rdf.util.PowerModel;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import com.interdataworking.mm.interpreter.*;


public class ReverseNonIdentical implements UntypedGateway {

  Set replaced(Model m, List map, boolean inverse) throws ModelException {

    Set s = new HashSet();
    NodeFactory f = m.getNodeFactory();
    Map l2r = MapPair.toJavaSetNodeMap(map, inverse);
    
    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement st = (Statement)en.nextElement();
      RDFUtil.replaceMult(st, f, l2r, s);
    }
    return s;
  }

  void findNonId(Model m1, Model m2, List map, Set nonId, Set unmatched, boolean inverse) throws ModelException {

    Set s = replaced(m1, map, inverse);

    for(Iterator it = s.iterator(); it.hasNext();) {
      
      Statement st = (Statement)it.next();
      if(!m2.contains(st)) {
	if(st.object() instanceof Literal)
	  nonId.add(st.subject()); // different literal
// 	else
//  	  unmatched.add(st.object()); // new/different/missing parent
      }
    }
  }

  public List reverseNonIdentical(Model m1, Model m2, List map) throws ModelException {

    Set nonId = new HashSet();
    Set unmatched = new HashSet();

    findNonId(m1, m2, map, nonId, unmatched, false);
    findNonId(m2, m1, map, nonId, unmatched, true);

    List result = new ArrayList();

    for(Iterator it = map.iterator(); it.hasNext();) {

      MapPair p = (MapPair)it.next();

//       if(unmatched.contains(p.getLeft()) || unmatched.contains(p.getRight()))
// 	continue; // do not add pair

      p = p.duplicate();
      
      if(nonId.contains(p.getLeft()) || nonId.contains(p.getRight()))
	p.inverse = !p.inverse;
      result.add(p);
    }

    return result;
  }


  public List execute(List input) throws ModelException {

    Model left = (Model)input.get(0);
    Model right = (Model)input.get(1);
    Model mapModel = (Model)input.get(2);

    List product = MapPair.toMapPairs(mapModel);
    List res = reverseNonIdentical(left, right, product);

    ArrayList l = new ArrayList();
    l.add(MapPair.asModel(mapModel, res));
    return l;
  }
  

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 3; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }

}
