package com.interdataworking.mm.gui;

import javax.swing.*;
import javax.swing.tree.*;
//import javax.swing.event.*;
//import javax.swing.text.*;
//import javax.swing.border.*;
//import javax.swing.colorchooser.*;
//import javax.swing.filechooser.*;
import javax.swing.table.*;
import javax.swing.event.ChangeEvent;
import javax.accessibility.*;

import java.awt.Event;
import java.awt.Image;
import java.awt.Graphics;
import java.awt.BorderLayout;
import java.awt.Point;
import java.awt.Color;
import java.awt.Container;
import java.awt.Component;
import java.awt.Insets;
import java.awt.event.*;
import java.awt.geom.*;
import java.beans.*;
import java.util.*;
import java.io.*;
import java.applet.*;
import java.net.*;

import com.interdataworking.*;
import com.interdataworking.mm.*;
import com.interdataworking.mm.alg.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;

public class AdjustSelector implements UntypedGateway {

  boolean isModal = true; // whether JDialog is modal
  Object lock = new Object(); // lock for modal

  AdjustMapping.MapPane[] panes = new AdjustMapping.MapPane[1]; // single pane only
  Set selector;

  /**
   * main method allows us to run as a standalone demo.
   */
  public static void main(String[] args) throws Exception {

    RDFFactory rf = new RDFFactoryImpl();
    NodeFactory f = rf.getNodeFactory();

    String ltext = "This is \n an example\n of some\ntext";

    Map llines = new HashMap();

    llines.put(f.createResource("1"), new TLine(1, 0, 4));
    llines.put(f.createResource("2"), new TLine(2, 4, 6));
    llines.put(f.createResource("3"), new TLine(3, 3, 6));
    
    Set selector = new HashSet();
    selector.add(f.createResource("1"));

    AdjustSelector rm = new AdjustSelector();
    rm.adjustSelector(ltext, llines, selector, "Sample selector");

    System.err.println("Done: " + selector);
  }

  public Set getSelector() {

    return selector;
  }

  public void adjustSelector(String ltext, Map llines, Set selector, String windowTitle) throws ModelException {

    JFrame frame = new JFrame();
    JDialog dialog = new JDialog(frame, "Selector Editor" + (windowTitle != null ? ": " + windowTitle : ""), false);
    /* {

	public boolean handleEvent(Event ev) {

	  if(ev.id == Event.WINDOW_DESTROY) {
	    this.dispose();
	    return true;
	  }
	  return super.handleEvent(ev);
	}
	};*/
    // non-modal in the AWT sense (blocking input to other windows)

    dialog.addWindowListener(new AdjustMapping.AMWindowAdapter(this.lock));

    this.selector = selector != null ? selector : new HashSet();

    panes[0] = AdjustMapping.createPane();
    panes[0].mouseListener = new ASMouseAdapter();

    BoxUpdater lbu = new ASBoxUpdater();

    panes[0].panel.add(new ImageComponent(ltext, llines, lbu));
    JScrollPane lscroll = new JScrollPane(panes[0].panel);

    dialog.getContentPane().add(lscroll);

    dialog.setSize(400, 600); // dialog.getHeight());
    dialog.setLocation(100, 100);

    dialog.setVisible(true);

    if(isModal) { // block until notified
      AdjustMapping.waitOnLock(this.lock);
    }
  }


  class ASMouseAdapter extends MouseAdapter {

    public void mouseEntered(MouseEvent e) {
      //	  System.err.println("Mouse entered");
      
      PointerBox p = (PointerBox)e.getComponent();
      p.mouseOver = true;
      //	  p.invalidate();
      panes[0].panel.repaint();
    }
    
    public void mouseExited(MouseEvent e) {
      //	  System.err.println("Mouse exited");
      PointerBox p = (PointerBox)e.getComponent();
      p.mouseOver = false;
      //	  p.invalidate();
      panes[0].panel.repaint();
    }
    
    public void mouseClicked(MouseEvent e) {
      
      PointerBox p = (PointerBox)e.getComponent();

      Object n = panes[0].getNode(p);

      if(p.selected) {
	selector.remove(n);
	p.selected = false;
      } else {
	selector.add(n);
	p.selected = true;
      }

      //	setMappings();

      panes[0].panel.repaint();

      //	  System.err.println("Clicked " + e.getClickCount() + " at " + e.getX() + ", " + e.getY() + " : " + p.hasFocus());
    }
  }

  class ASBoxUpdater implements BoxUpdater {

    public void updateBoxes(Map boxes, Object src) {

      AdjustMapping.MapPane pane = panes[0];

      for(Iterator it = boxes.entrySet().iterator(); it.hasNext();) {

	Map.Entry e = (Map.Entry)it.next();
	JComponent box = (JComponent)e.getValue();
	RDFNode n = (RDFNode)e.getKey();

	pane.panel.add(box, 0);
	box.addMouseListener(pane.mouseListener);

	if(selector.contains(n))
	  ((PointerBox)box).selected = true;
      }
      pane.r2j = boxes;
      pane.j2r = AdjustMapping.invert(boxes);

      pane.panel.repaint();
      //      setMappings();
    }
  }


  /**
   * Parameters:
   * (string1, lines1, selector, winTitle)
   * Returns:
   *  map
   */
  public List execute(List input) throws ModelException {

    String ltext = (String)input.get(0);
    Map llines = (Map)input.get(1);

    Model setM = (Model)input.get(2);
    Set set = MapVocabulary.asSet(setM);

    String windowTitle = null;
    if(input.size() > 3)
      windowTitle = (String)input.get(3);

    adjustSelector(ltext, llines, set, windowTitle);

    List result = new ArrayList();
    Model r = setM == null ? new RDFFactoryImpl().createModel() : setM.create();
    result.add(MapVocabulary.asModel(r, getSelector()));

    return result;
  }

  public int getMinInputLen() { return 3; }

  public int getMaxInputLen() { return 4; }

  public int getMinOutputLen() { return 1; }

  public int getMaxOutputLen() { return 1; }
}
