package com.interdataworking.util;

import java.util.*;
import com.interdataworking.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import org.w3c.rdf.implementation.model.*;
import edu.stanford.db.rdf.query.util.*;


/**
 */

public abstract class SimpleModelGateway implements ModelGateway {

  static NodeFactory defaultNodeFactory = new NodeFactoryImpl();

  protected Model dest;
  protected Model src;

  int currentPass = 0;
  boolean anotherPass;

  HashMap skolems; // skolem -> resource

  Collection cleanUpTemplates = new ArrayList(); // of StatementTemplates

  protected abstract void processStatement(Statement st) throws ModelException;

  protected Resource getUnique(Skolem s) throws ModelException {

    Resource r = (Resource)skolems.get(s);
    if(r == null) {
      r = dest.getNodeFactory().createUniqueResource();
      skolems.put(s, r);
    }
    return r;
  }

  protected void setCleanUpTemplates(Collection c) {

    cleanUpTemplates = c;
  }

  protected void addCleanUp(Resource s, Resource p, RDFNode o) {

    if(cleanUpTemplates == null)
      cleanUpTemplates = new ArrayList();
    StatementTemplate t = new StatementTemplate(s, p, o);
    cleanUpTemplates.add(t);
    //    System.err.println("New template: " + t + " of " + cleanUpTemplates.size());
  }

  public void init(Model srcMD, Model destMD) throws ModelException {

    cleanUpTemplates = new ArrayList();
  }  

  protected void doAnotherPass() {

    anotherPass = true;
  }

  public Model translate(Model src) throws ModelException {

    skolems = new HashMap();
    this.src = src;
    dest = src.duplicate();

    currentPass = 0;
    anotherPass = false;

    while(++currentPass <= totalPassNum() || anotherPass) {
      
      //      System.err.println("PASS #" + currentPass + " for " + dest.size() + " statements from model impl " + dest.getClass());

      Enumeration en = (currentPass == 1 ? src.elements() : dest.duplicate().elements());
      while(en.hasMoreElements()) {
	
	Statement st = (Statement)en.nextElement();

	//	System.err.println(" " + st.hashCode() + " " + st.getClass());

	if(st != null /* TODO: bug?? */ && dest.contains(st)) { // ignore statements that have already been discarded
// 	if(true)
// 	  continue;

	  //	  System.out.println("DEST SIZE BEFORE: " + dest.size());
	  processStatement(st);
	  //	  System.out.println("DEST SIZE AFTER: " + dest.size());
	}
      }
      anotherPass = false;
    }

    cleanUp();
    return dest;
  }

  protected void cleanUp() throws ModelException {

    // clean up if necessary
    //    System.err.println("Cleaning up: " + cleanUpTemplates.size());
    if(cleanUpTemplates != null) {    
      Iterator it = cleanUpTemplates.iterator();
      while(it.hasNext()) {
	StatementTemplate t = (StatementTemplate)it.next();
	Model mt = t.findIn(dest);
	//	System.err.println("FOUND " + mt + " FOR " + t);
	SetOperations.subtract(dest, mt);
      }
    }
  }

  protected int currentPass() {

    return currentPass;
  }

  protected int totalPassNum() {

    return 1;
  }

  protected boolean contains(Resource s, Resource p, RDFNode o) throws ModelException {

    return dest.find(s, p, o).size() != 0;
  }

  protected Statement get(RDFNode s, RDFNode p, RDFNode o) throws ModelException {

    if(!(s instanceof Resource) || !(p instanceof Resource))
      return null;

    return RDFUtil.get1(dest.find((Resource)s, (Resource)p, o));
  }

  protected Statement stmt(Resource s, Resource p, RDFNode o) throws ModelException {

    return dest.getNodeFactory().createStatement(s, p, o);
  }

  protected NodeFactory factory() throws ModelException {

    return dest != null ? dest.getNodeFactory() : defaultNodeFactory;
  }

}
