package edu.stanford.db.alg.ngram;

import java.util.*;

public class ElementSimilarity {

  public static final double PRECISION = 0.0000001;

  Object e1, e2;
  double sim;

  public ElementSimilarity(Object e1, Object e2) {

    this.e1 = e1;
    this.e2 = e2;
  }

  public ElementSimilarity(Object e1, Object e2, double sim) {

    this(e1, e2);
    this.sim = sim;
  }

  private ElementSimilarity() {
  }

  public Object getElement1() {

    return e1;
  }

  public Object getElement2() {

    return e2;
  }

  public double getSim() {

    return sim;
  }

  public void setSim(double sim) {

    this.sim = sim;
  }

  public void incSim(double sim) {

    this.sim += sim;
  }

  public int hashCode() {

    return e1.hashCode() * 37 + e2.hashCode();
  }

  public boolean equals(Object that) {

    if(this == that)
      return true;

    if(!(that instanceof ElementSimilarity))
      return false;

    ElementSimilarity t = (ElementSimilarity)that;

    return e1.equals(t.e1) && e2.equals(t.e2);
  }

  public String toString() {

    return "Similarity[" + sim + ": " + e1 + ", " + e2 + "]";
  }

  public static ElementSimilarity createContainer() {

    return new ElementSimilarity();
  }

  public static ElementSimilarity get(Map map, Object e1, Object e2, ElementSimilarity container) {

    container.e1 = e1;
    container.e2 = e2;
    ElementSimilarity so = (ElementSimilarity)map.get(container);
    if(so == null) {
      so = new ElementSimilarity(e1, e2);
      map.put(so, so);
    } 
    return so;
  }

  /** sort an array of map pairs */
  public static void sort(Object[] arr) {
    
    Arrays.sort(arr, new Comparator() {
	
	public int compare(Object o1, Object o2) {
	  ElementSimilarity p1 = (ElementSimilarity)o1;
	  ElementSimilarity p2 = (ElementSimilarity)o2;
	  double diff = p1.sim - p2.sim;
	  if(Math.abs(diff) < PRECISION) { // consider equal
	    // be deterministic
	    int h1 = p1.hashCode();
	    int h2 = p2.hashCode();
	    if(h1 < h2)
	      return -1;
	    else if(h1 > h2)
	      return 1;
	    else
	      return 0;
	  } else if(diff > 0)
	    return -1;
	  else //if(diff < 0)
	    return 1;
	}

	public boolean equals(Object that) {
	  return this == that;
	}
      });
  }

}
