/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;

import java.util.*;

/**
 * This is a general-purpose condition iterator.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class ConditionIterator implements QueryIterator {

  QueryIterator it;
  /** ordered array of variable names */
  String[] var;
  /** holds a runtime variable substitution */
  RDFNode[] substitution;
  ConditionEvaluator evaluator;

  public ConditionIterator(String var, ConditionEvaluator evaluator, QueryIterator it) {

    this(oneStringArray(var), evaluator, it);
  }

  public ConditionIterator(String var1, String var2, ConditionEvaluator evaluator, QueryIterator it) {

    this(twoStringArray(var1, var2), evaluator, it);
  }

  public ConditionIterator(String[] var, ConditionEvaluator evaluator, QueryIterator it) {
    
    this.var = var;
    this.it = it;
    this.evaluator = evaluator;
    if(evaluator.getArity() != var.length)
      throw new RuntimeException("Number of parameters incompatible with arity of condition evaluator " + evaluator);

    this.substitution = new RDFNode[var.length];
  }

  public static String[] oneStringArray(String str) {

    String[] names = new String[1];
    names[0] = str;
    return names;
  }

  public static String[] twoStringArray(String str1, String str2) {

    String[] names = new String[2];
    names[0] = str1;
    names[1] = str2;
    return names;
  }

  // resets iterator with the preset variables. If model passed,
  // resets this and underlying iterators with its value
  public void reset(Hashtable vars, Model model) throws ModelException {

    it.reset(vars, model);
  }

  // returns true if match found (vars updated if necessary)
  public boolean next(Hashtable vars) throws ModelException {

    while(it.next(vars)) {

      for(int i=0; i < var.length; i++)
	substitution[i] = (RDFNode)vars.get(var[i]);

      if(evaluator.holds(substitution))
	return true;
    }
    return false;
  }

  // returns names of the variables used
  public Enumeration getVariables() throws ModelException {

    return it.getVariables();
  }

  // returns execution cost of the iterator (e.g. stylizised number
  // of instructions/method calls)
  public int getCost() {

    return it.getCost();
  }

  public QueryIterator duplicate() {
    
    // keep same evaluator and variable names
    return new ConditionIterator(var, evaluator, it.duplicate());
  }

}
