/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;

import java.util.*;

/**
 * This is a query iterator that implements negation.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class NegatedIterator implements QueryIterator {

  boolean noMore = false;
  QueryIterator it;

  public NegatedIterator(QueryIterator it) {

    this.it = it;
  }

  // resets iterator with the preset variables. If model passed,
  // resets this and underlying iterators with its value
  public void reset(Hashtable vars, Model model) throws ModelException {

    it.reset(vars, model);
    noMore = false;
  }

  // returns true if match found (vars updated if necessary)
  public boolean next(Hashtable vars) throws ModelException {

    if(noMore)
      return false;

    noMore = true;
    Hashtable t = (Hashtable)vars.clone();
    boolean b = it.next(t);
    return !b;
  }

  // returns names of the variables used
  public Enumeration getVariables() throws ModelException {

    return it.getVariables();
  }

  // returns execution cost of the iterator (e.g. stylizised number
  // of instructions/method calls)
  public int getCost() {

    return it.getCost();
  }

  public QueryIterator duplicate() {

    NegatedIterator n = new NegatedIterator(it.duplicate());
    n.noMore = noMore;
    return n;
  }

}
