/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.implementation.model.ResourceImpl;
import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.util.RDFUtil;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import edu.stanford.db.rdf.query.util.Skolem;

import java.util.*;

/**
 * This is a query iterator that sets a flat whether a given
 * variable is seen for the first time
 * If flagVarName is not set, repetitions are tacitly skipped,
 * otherwise variable flagVarName will be set to UNIQUE or SKIP
 *
 * @deprecated use GroupBy operator
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class UniqueIterator implements QueryIterator {

  public static final Resource UNIQUE = new ResourceImpl("UniqueIterator:unique");
  public static final Resource SKIP = new ResourceImpl("UniqueIterator:skip");

  QueryIterator it;
  String uniqueVarName;
  String flagVarName;
  HashSet doneValues = new HashSet();

  public UniqueIterator(String uniqueVarName, QueryIterator it) {

    this(it, uniqueVarName, null);
  }

  /**
   * @deprecated constructor
   */
  public UniqueIterator(QueryIterator it, String uniqueVarName) {

    this(it, uniqueVarName, null);
  }

  /**
   * @deprecated constructor
   */
  public UniqueIterator(QueryIterator it, String uniqueVarName, String flagVarName) {

    this.it = it;
    this.uniqueVarName = uniqueVarName;
    this.flagVarName = flagVarName;
  }

  public UniqueIterator(String uniqueVarName, String flagVarName, QueryIterator it) {

    this.it = it;
    this.uniqueVarName = uniqueVarName;
    this.flagVarName = flagVarName;
  }

  public void reset(Hashtable vars, Model model) throws ModelException {

    it.reset(vars, model);
    doneValues = new HashSet();
  }

  // returns true if match found (vars updated if necessary)
  public boolean next(Hashtable vars) throws ModelException {

    boolean tryAgain = true;

    do {
      if(!it.next(vars))
	return false;
      Object o = vars.get(uniqueVarName);
      if(o == null)
	return true;
      boolean inDoneValues = doneValues.contains(o);
      tryAgain = inDoneValues && flagVarName == null;
      doneValues.add(o);
      if(flagVarName != null)
	vars.put(flagVarName, inDoneValues ? SKIP : UNIQUE);
    } while(tryAgain);

    return true;
  }

  // returns names of the variables used
  public Enumeration getVariables() throws ModelException {

    return it.getVariables();
  }

  // returns execution cost of the iterator (e.g. stylizised number
  // of instructions/method calls)
  public int getCost() {

    return it.getCost();
  }

  public QueryIterator duplicate() {

    UniqueIterator oi = new UniqueIterator(it.duplicate(), uniqueVarName);
    oi.doneValues = (HashSet)doneValues.clone();
    return oi;
  }

}
