package edu.stanford.db.rdf.query.orderings;

import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;


/**
 * Assumes that there is an ordering property on statements,
 * whose value is used for comparisons.
 * The statements used in comparisons are identified using variable named <tt>stmtVar</tt>.
 */

public class ByOrderProperty implements SubstitutionComparator {

  String stmtVar;
  Resource orderResource;

  public ByOrderProperty(String stmtVar) {

    this(stmtVar, RDFX.order);
  }

  public ByOrderProperty(String stmtVar, Resource orderResource) {

    this.stmtVar = stmtVar;
    this.orderResource = orderResource;
  }

  public int compare(Map t1, Map t2, Model m) throws ModelException {

    if(m == null)
      throw new ModelException("[ByOrderProperty] call reset to initialize the model first!");

    // System.err.println("==== comparing " + t1 + " and " + t2);

    Statement s1 = (Statement)t1.get(stmtVar);
    Statement s2 = (Statement)t2.get(stmtVar);

    // obtain order of the statement
    Literal l1 = (Literal)RDFUtil.getObject(m, s1, orderResource);
    Literal l2 = (Literal)RDFUtil.getObject(m, s2, orderResource);
    
    // 			  if(l1 == null)
    // 			    System.err.println("MISSING ORDER 1: " + s1 + ", order var: " + orderResource);

    // 			  if(l2 == null)
    // 			    System.err.println("MISSING ORDER 2: " + s2);

    if(l1 == null && l1 != null) // stuff w/o order goes last
      return 1;
    else if(l2 == null && l1 != null)
      return -1;
    else if(l1 == null && l2 == null)
      return t1.hashCode() - t2.hashCode(); // be deterministic

    // else: both are set, convert them to integers
    try {

      int i1 = Integer.parseInt(l1.getLabel());
      int i2 = Integer.parseInt(l2.getLabel());

      if(i1 < i2)
	return -1;
      else if(i1 > i2)
	return 1;
      return 0;

    } catch (Exception any) {

      return 0;
    }
  }
}

