package edu.stanford.db.rdf.util;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;

public class Aggregator {

  static final Collection EMPTY = new HashSet();
  static boolean DEBUG = true;

  Model m;
  Collection aggProp, inverseAggProp, weakAggProp, weakInverseAggProp;
  Set result = new HashSet();
  //  Set visitedNodes = new HashSet();
  Set ignorableEdges = new HashSet();

  // bookkeeping only
  Set referencedNodes = new HashSet();
  Set pulledOutNodes = new HashSet();

  public Aggregator(Collection aggProp, Collection inverseAggProp, Collection weakAggProp, Collection weakInverseAggProp, Model m) {

    this.aggProp = aggProp != null ? aggProp : EMPTY;
    this.inverseAggProp = inverseAggProp != null ? inverseAggProp : EMPTY;
    this.weakAggProp = weakAggProp != null ? weakAggProp : EMPTY;
    this.weakInverseAggProp = weakInverseAggProp != null ? weakInverseAggProp : EMPTY;
    this.m = m;
  }

  public Aggregator(Resource[] aggProp, Resource[] inverseAggProp, Resource[] weakAggProp, Resource[] weakInverseAggProp, Model m) {

    Set s1 = new HashSet();
    s1.addAll(Arrays.asList(aggProp));

    Set s2 = new HashSet();
    s2.addAll(Arrays.asList(inverseAggProp));

    Set s3 = new HashSet();
    s3.addAll(Arrays.asList(weakAggProp));

    Set s4 = new HashSet();
    s4.addAll(Arrays.asList(weakInverseAggProp));

    this.aggProp = s1;
    this.inverseAggProp = s2;
    this.weakAggProp = s3;
    this.weakInverseAggProp = s4;
    this.m = m;
  }

  /*
  public Set getVisitedNodes() {

    return visitedNodes;
  }

  public void setVisitedNodes(Set v) {

    visitedNodes = v;
  }
  */

  /** returns nodes that have been referenced only, but not pulled out.
   * referencedNodes is used for bookkeeping only */

  public Set getReferencedNodes() {

    return referencedNodes;
  }

  public void setReferencedNodes(Set v) {

    referencedNodes = v;
  }

  /** returns nodes that have been pulled out */

  public Set getPulledOutNodes() {

    return pulledOutNodes;
  }

  public void setPulledOutNodes(Set v) {

    pulledOutNodes = v;
  }

  public Set getIgnorableEdges() {

    return ignorableEdges;
  }

  public void setIgnorableEdges(Set i) {

    ignorableEdges = i;
  }

  public Set getResult() {

    return result;
  }

  public void clearResult() {

    result.clear();
    referencedNodes.clear();
  }

  public void pullOutNode(RDFNode n) throws ModelException {

    pullOutNode(n, 0);
  }
  
  // returns true of pulled out something

  void pullOutNode(RDFNode n, int depth) throws ModelException {
    
    if(n == null || pulledOutNodes.contains(n))
      return;

    if(DEBUG && !(n instanceof Statement)) {
//       for(int i=0; i < depth; i++)
// 	System.err.print("\t");
//       System.err.println("pulling out: " + n);
    }

    pulledOutNodes.add(n);
    referencedNodes.remove(n);

    // aggProp
    if(n instanceof Resource) {
      for(Enumeration en = m.find((Resource)n, null, null).elements(); en.hasMoreElements(); ) {
	
	Statement st = (Statement)en.nextElement();
	Resource pred = st.predicate();
	
	if(ignorableEdges.contains(st))
	  continue;

	boolean byValue = /*depth == 0 || */ aggProp.contains(pred);
	boolean byRef = /*depth == 0 || */ weakAggProp.contains(pred);

	if(byValue) {

	  result.add(st);
	  // check for reified statements
	  pullOutNode(st, depth+1); //0);
	  // need to pull out/examine object
	  pullOutNode(st.object(), depth+1);

	} else if(byRef) {

	  result.add(st);
	  // check for reified statements
	  pullOutNode(st, depth+1); //0);
	  // record reference that has not been pulled out
	  if(!pulledOutNodes.contains(st.object()))
	    referencedNodes.add(st.object());
	}

      }
    }

    // inverseAggProp
    for(Enumeration en = m.find(null, null, n).elements(); en.hasMoreElements(); ) {

      Statement st = (Statement)en.nextElement();
      Resource pred = st.predicate();

      if(ignorableEdges.contains(st))
	continue;

      boolean byValue = /*depth == 0 ||*/ inverseAggProp.contains(pred);
      boolean byRef = /*depth == 0 ||*/ weakInverseAggProp.contains(pred);

	if(byValue) {
	  result.add(st);
	  // check for reified statements
	  pullOutNode(st, depth+1); //0);
	  // need to pull out/examine subject
	  pullOutNode(st.subject(), depth+1);

	} else if(byRef) {

	  result.add(st);
	  // check for reified statements
	  pullOutNode(st, depth+1); //0);
	  // record reference that has not been pulled out
	  if(!pulledOutNodes.contains(st.subject()))
	    referencedNodes.add(st.subject());
	}
    }
  }
}
