/**
 * Copyright  World Wide Web Consortium, (Massachusetts Institute of
 * Technology, Institut National de Recherche en Informatique et en
 * Automatique, Keio University).
 *
 * All Rights Reserved.
 *
 * Please see the full Copyright clause at
 * <http://www.w3.org/Consortium/Legal/copyright-software.html>
 *
 */

package org.w3c.rdf.model;

/**
 * An RDF resource. Every RDF resource must have a URI.  Many RDF
 * resources are pieces of vocabulary. They typically have a namespace
 * and a local name. In this case, a URI is composed as a
 * concatenation of the namespace and the local name.
 * For other kinds of identifiers (e.g. autogenerated UUIDs), the namespace is null.
 */

public interface Resource extends RDFNode {

  /**
   * Returns the URI of the resource. Triples and models must implement this method in a standard way.
   * @return the URI of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   */
  public String getURI() throws ModelException;

  /**
   * Returns the namespace of the resource. May return null.
   * @return the namespace of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   * @since 2000-10-21
   */
  public String getNamespace() throws ModelException;

  /**
   * Returns the local name of the resource. May not return null.
   * @return the local name of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   * @since 2000-10-21
   */
  public String getLocalName() throws ModelException;
}

